# -*- mode: python ; coding: utf-8 -*-
#
# Python-slip39 -- Ethereum SLIP-39 Account Generation and Recovery
#
# Copyright (c) 2022, Dominion Research & Development Corp.
#
# Python-slip39 is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.  It is also available under alternative (eg. Commercial) licenses, at
# your option.  See the LICENSE file at the top of the source tree.
#
# Python-slip39 is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#

from __future__		import annotations

import logging
import math

from collections	import namedtuple
from typing		import Tuple, Callable, Type, Optional

from fpdf		import FPDF

from ..util		import ordinal
from ..defaults		import (
    FONTS, MNEM_ROWS_COLS, MM_IN, PT_IN, COLOR, PAGE_MARGIN,
)

log				= logging.getLogger( __package__ )


class Region:
    """Takes authority for a portion of another Region, and places things in it, relative to its
    upper-left and lower-right corners.

    Attempts to create a sane priority hierarchy, because fpdf2's FlexTemplate sorts based on it.
    If a specific priority is provided, it will be used.  However, if no priority is given, 0 will
    be automatically assigned by FlexTemplate.  If sorting is unstable, this may result in an
    unspecified ordering.

    """
    def __init__( self, name, x1=None, y1=None, x2=None, y2=None, rotate=None, priority=None ):
        self.name		= name
        self.x1			= x1		# could represent positions, offsets or ratios
        self.y1			= y1
        self.x2			= x2
        self.y2			= y2
        self.rotate		= rotate
        self.priority		= priority
        self.regions		= []

    @property
    def h( self ):
        return self.y2 - self.y1

    @property
    def w( self ):
        return self.x2 - self.x1

    def element( self ):
        "Converts to mm.  Optionally returns a specified priority."
        d			= dict(
            name	= self.name,
            x1		= self.x1 * MM_IN,
            y1		= self.y1 * MM_IN,
            x2		= self.x2 * MM_IN,
            y2		= self.y2 * MM_IN,
            rotate	= self.rotate or 0.0
        )
        if self.priority:
            d['priority']	= self.priority
        return d

    def __str__( self ):
        return f"({self.name:16}: ({float(self.x1):8}, {float( self.y1):8}) - ({float(self.x2):8} - {float(self.y2):8})"

    def add_region( self, region ):
        region.x1		= self.x1 if region.x1 is None else region.x1
        region.y1		= self.y1 if region.y1 is None else region.y1
        region.x2		= self.x2 if region.x2 is None else region.x2
        region.y2		= self.y2 if region.y2 is None else region.y2
        self.regions.append( region )
        return region

    def add_region_relative( self, region ):
        region.x1		= self.x1 + ( region.x1 or 0 )
        region.y1		= self.y1 + ( region.y1 or 0 )
        region.x2		= self.x2 + ( region.x2 or 0 )
        region.y2		= self.y2 + ( region.y2 or 0 )
        self.regions.append( region )
        return region

    def add_region_proportional( self, region ):
        region.x1		= self.x1 + self.w * ( 0 if region.x1 is None else region.x1 )
        region.y1		= self.y1 + self.h * ( 0 if region.y1 is None else region.y1 )
        region.x2		= self.x1 + self.w * ( 1 if region.x2 is None else region.x2 )
        region.y2		= self.y1 + self.h * ( 1 if region.y2 is None else region.y2 )
        self.regions.append( region )
        return region

    def square( self, maximum=None, justify=None ):
        """Make a region square (of 'maximum' size), justifying it L/C/R, and/or T/M/B; default is Center/Middle"""
        dims			= min( self.w, self.h )
        if maximum:
            dims		= min( dims, maximum )
        justify			= ( justify or '' ).upper()
        if 'L' in justify:
            self.x2		= self.x1 + dims
        elif 'R' in justify:
            self.x1		= self.x2 - dims
        else:  # 'C' is default
            self.x1, self.x2	= self.x1 + ( self.w - dims ) / 2, self.x2 - ( self.w - dims ) / 2
        if 'T' in justify:
            self.y2		= self.y1 + dims
        elif 'B' in justify:
            self.y1		= self.y2 - dims
        else:  # 'M' is default
            self.y1, self.y2	= self.y1 + ( self.h - dims ) / 2, self.y2 - ( self.h - dims ) / 2
        return self

    def mm( self ):
        "Dimensions as Coordinate( x, y ) in mm."
        return Coordinate( self.w * MM_IN, self.h * MM_IN )

    dimensions 			= mm

    def elements( self ):
        """Yield a sequence of { 'name': "...", 'x1': #,  ... }."""
        if self.__class__ != Region:
            yield self.element()
        for r in self.regions:
            for d in r.elements():
                yield d


class Text( Region ):
    SIZE_RATIO			= 3/4

    def __init__( self, *args, font=None, style=None, text=None, size=None, size_ratio=None, align=None, multiline=None,
                  foreground=None, background=None, bold=None, italic=None, underline=None,
                  **kwds ):
        self.font		= font		# "dejavu"
        self.style		= style		# eg. "BI".  Or, built (in fpdf) from self.bold, self.italic
        self.text		= text
        self.multiline		= multiline
        self.size		= size
        self.size_ratio		= size_ratio or self.SIZE_RATIO
        self.align		= align
        self.foreground		= foreground
        self.background		= background
        self.bold		= bold
        self.italic		= italic
        self.underline		= underline
        super().__init__( *args, **kwds )

    def element( self ):
        d			= super().element()
        d['type']		= 'T'
        d['font']		= FONTS.get( self.font ) or self.font or FONTS.get( 'sans' )
        if self.style:
            d['style']		= self.style
        line_height		= self.h * PT_IN  # Postscript point == 1/72 inch
        d['size']		= self.size or ( line_height * self.size_ratio )  # No need for int(round(..))
        if self.text is not None:
            d['text']		= self.text
        if self.bold:
            d['bold']		= True
        if self.italic:
            d['italic']		= True
        if self.underline:
            d['underline']	= True
        if self.align is not None:
            d['align']		= self.align
        if self.multiline is not None:
            d['multiline']	= bool( self.multiline )
        if self.foreground is not None:
            d['foreground']	= self.foreground
        if self.background is not None:
            d['background']	= self.background
        return d


class Image( Region ):
    def __init__( self, *args, font=None, text=None, size=None, **kwds ):
        self.font		= font
        self.text		= text
        self.size		= None
        super().__init__( *args, **kwds )

    def element( self ):
        d			= super().element()
        d['type']		= 'I'
        return d


class Line( Region ):
    def emit( self, d, *args, **kwds ):
        d['type']		= 'L'
        super().emit( *args, **kwds )


class Box( Region ):
    def element( self ):
        d			= super().element()
        d['type']		= 'B'
        return d


Coordinate			= namedtuple( 'Coordinate', ('x', 'y') )


def layout_card(
    card_size: Coordinate,
    card_margin: int,
    num_mnemonics: int	= 20,
):
    card			= Box( 'card', 0, 0, card_size.x, card_size.y )
    card_interior		= card.add_region_relative(
        Region( 'card-interior', x1=+card_margin, y1=+card_margin, x2=-card_margin, y2=-card_margin )
    )

    # Rotate the card group, watermark, etc. so its angle is from the lower-left to the upper-right.
    #
    #             b
    #      +--------------+
    #      |           .
    #    a |       .
    #      | β .    c
    #      + 90-β
    #
    #   tan(β) = b / a, so
    #   β = arctan(b / a)
    #
    a				= card_interior.h
    b				= card_interior.w
    c				= math.sqrt( a * a + b * b )
    β				= math.atan( b / a )
    rotate			= 90 - math.degrees( β )

    for c_n in range( 16 ):  # SLIP-39 supports up to 16 groups
        card_interior.add_region_proportional( Text(
            f'card-g{c_n}', x1=c/b * 8/100, y1=-2/32, x2=c/b * 92/100, y2=10/32,
            foreground	= int( COLOR[c_n % len( COLOR )], 16 ),
            rotate	= -rotate,
            bold	= True,
        ))
    card_interior.add_region_proportional( Text(
        'card-watermark', x1=5/100, y1=14/16, x2=c/b * 95/100, y2=17/16,
        foreground	= int( COLOR[-2], 16 ),
        rotate		= rotate,  # eg 60
        align		= 'L'
    ))

    # Rotation is downward around upper-left corner; so, lower-left corner will shift 1 height
    # leftward and upward; so start 1 height right and down.
    link_length			= card_interior.h
    link_height			= link_length * 6/32
    card_interior.add_region(
        Text(
            'card-link',
            x1		= card_interior.x1 + link_height,
            y1		= card_interior.y1,
            x2		= card_interior.x1 + link_height + link_length,
            y2		= card_interior.y1 + link_height,
            rotate	= -90,				# along left border of card
            foreground	= int( COLOR[-3], 16 ), 	# Light grey
            align	= 'R'
        )
    )

    card_top			= card_interior.add_region_proportional(
        Region( 'card-top', x1=0, y1=0, x2=1, y2=1/4 )
    )
    card_bottom			= card_interior.add_region_proportional(
        Region( 'card-bottom', x1=0, y1=1/4, x2=1, y2=1 )
    )
    card_mnemonics		= card_bottom.add_region_proportional(
        Region( 'card-mnemonics', x1=0, y1=0, x2=13/16, y2=1 )
    )
    # QR code for Card Mnemonics on back is center/middle of Mnemonics area
    card_mnemonics.add_region(
        Image( 'card-qr-mnem' )
    ).square( justify='T' )

    # QR codes sqaare, and anchored to top and bottom of card.
    card_bottom.add_region_proportional(
        Image( 'card-qr1', x1=13/16, y1=0, x2=1, y2=1/2 )
    ).square( justify='TR' )
    card_bottom.add_region_proportional(
        Image( 'card-qr2', x1=13/16, y1=1/2, x2=1, y2=1 )
    ).square( justify='BR' )

    card_top.add_region_proportional(
        Text( 'card-title', x1=0, y1=0, x2=1, y2=40/100, bold=True )
    )
    card_top.add_region_proportional(
        Text( 'card-requires', x1=0, y1=40/100, x2=1, y2=66/100, align='C', italic=True )
    )
    card_top.add_region_proportional(
        Text( 'card-crypto1', x1=0, y1=66/100, x2=1, y2=83/100, align='R' )
    )
    card_top.add_region_proportional(
        Text( 'card-crypto2', x1=0, y1=83/100, x2=1, y2=100/100, align='R' )
    )

    assert num_mnemonics in MNEM_ROWS_COLS, \
        f"Invalid SLIP-39 mnemonic word count: {num_mnemonics}"
    rows,cols			= MNEM_ROWS_COLS[num_mnemonics]
    for r in range( rows ):
        for c in range( cols ):
            card_mnemonics.add_region_proportional(
                Text(
                    f"mnem-{c * rows + r}",
                    x1		= c/cols,
                    y1		= r/rows,
                    x2		= (c+1)/cols,
                    y2		= (r+1)/rows,
                    font	= 'mono',
                    bold	= True,
                    size_ratio	= 9/16,
                )
            )

    return card


def layout_wallet(
    wallet_size: Coordinate,
    wallet_margin: int,
):
    """Produce a template format for a Paper Wallet.

    Sets priority:
      -3     : Hindmost backgrounds
      -2     : Things atop background, but beneath contrast-enhancement
      -1     : Contrast-enhancing partially transparent images
       0     : Text, etc. (the default)
    """
    prio_backing		= -3
    prio_normal			= -2
    prio_contrast		= -1

    wallet			= Box( 'wallet', 0, 0, wallet_size.x, wallet_size.y )
    wallet_background		= wallet.add_region_proportional(  # noqa: F841
        Image( 'wallet-bg', priority=prio_backing ),
    )

    # Most of the Paper Wallet is visible (contains public address information).  The right-most
    # portion can be folded over, and displays private key information
    show			= wallet.add_region_proportional(
        Region( 'wallet-show', x2=5/8 )
    )
    fold			= wallet.add_region_proportional(
        Box( 'wallet-fold', x1=5/8 )
    )

    public			= show.add_region_relative(
        Region( 'wallet-public', x1=+wallet_margin, y1=+wallet_margin, x2=-wallet_margin, y2=-wallet_margin )
    )
    private			= fold.add_region_relative(
        Region( 'wallet-private', x1=+wallet_margin, y1=+wallet_margin, x2=-wallet_margin, y2=-wallet_margin )
    )

    # Assign each different Crypto name a different color, in template labels crypto-{f,b}1, crypto-{f,b}2, ...
    for c_n in range( len( COLOR )):
        public.add_region_proportional(
            Text( f'crypto-f{c_n}', x1=1/8, y1=-1/16, x2=7/8, y2=7/16, foreground=int( COLOR[c_n], 16 ), rotate=-45, priority=prio_normal )
        )
        private.add_region_proportional(
            Text( f'crypto-b{c_n}', x1=2/8, y1=-1/16, x2=7/8, y2=7/16, foreground=int( COLOR[c_n], 16 ), rotate=-45, priority=prio_normal )
        )

    # The background rosette and cryptocurrency symbol in the center
    public_center_size		= min( public.w, public.h )
    public.add_region(
        Image(
            'center',
            x1		= public.x1 + public.w * 2/3 - public_center_size / 3,
            y1		= public.y1 + public.h * 1/2 - public_center_size / 3,
            x2		= public.x1 + public.w * 2/3 + public_center_size / 3,
            y2		= public.y1 + public.h * 1/2 + public_center_size / 3,
        )
    )

    # Public addresses are vertical on left- and right-hand of public Region.  In order to fit the
    # longer ETH addresses into a space with a fixed-width font, we know that the ratio of the width
    # to the height has to be about 1/20.  Rotation is downward around upper-left corner; so,
    # lower-left corner will shift 1 height leftward and upward; so start 1 height right and down.
    address_length		= public.h
    address_height		= address_length * 1/20
    public.add_region(
        Image(
            'address-l-bg',
            x1		= public.x1 + address_height,
            y1		= public.y1,
            x2		= public.x1 + address_height + address_length,
            y2		= public.y1 + address_height,
            rotate	= -90,
            priority	= prio_contrast,
        )
    ).add_region(
        Text(
            'address-l',
            font	= 'mono',
            rotate	= -90,
        )
    )

    # Rotation is upward around the upper-left corner, so lower-left corner will shift 1 height
    # upward and right; so start 1 height leftward and down.
    public.add_region(
        Image(
            'address-r-bg',
            x1		= public.x2 - address_height,
            y1		= public.y2,
            x2		= public.x2 - address_height + address_length,
            y2		= public.y2 + address_height,
            rotate	= +90,
            priority	= prio_contrast,
        )
    ).add_region(
        Text(
            'address-r',
            font	= 'mono',
            rotate	= +90,
        )
    )

    # Wallet name, amount
    public.add_region_proportional(
        Text( 'name-label',	x1=1/16, y1=0/16, x2=4/16, y2=1/16 )
    )
    public.add_region_proportional(
        Image( 'name-bg',	x1=4/16, y1=0/16, x2=15/16, y2=1/16, priority=prio_contrast )
    )
    public.add_region_proportional(
        Text( 'name',		x1=4/16, y1=0/16, x2=15/16, y2=1/16 )
    )

    public.add_region_proportional(
        Text( 'amount-label',	x1=1/16, y1=3/16, x2=4/16, y2=1/16 )
    )
    public.add_region_proportional(
        Image( 'amount-bg',	x1=4/16, y1=0/16, x2=15/16, y2=1/16, priority=prio_contrast )
    )
    public.add_region_proportional(
        Text( 'amount',		x1=4/16, y1=0/16, x2=15/16, y2=1/16 )
    )

    # Make Public QR Code square w/ min of height, width, anchored at lower-left corner
    public.add_region_proportional(
        Text( 'address-qr-t',	x1=1/16, y1=5/16, x2=1, y2=6/16 )
    )
    public_qr			= public.add_region_proportional(
        Image( 'address-qr-bg',	x1=1/16, y1=6/16, x2=1, y2=15/16 )
    ).square( justify='BL' )
    public_qr.add_region(
        Image( 'address-qr' )
    )

    public.add_region_proportional(
        Text( 'address-qr-b',	x1=1/16, y1=15/16, x2=1, y2=16/16 )
    )
    public_qr_r			= public.add_region_proportional(
        Text( 'address-qr-r',	x1=1/16, y1=6/16, x2=1, y2=7/16, rotate=-90 )
    )
    public_qr_r.x1	       += public_qr.w + public_qr_r.h
    public_qr_r.x2	       += public_qr.w

    # Private region

    private.add_region_proportional(
        Text( 'private-qr-t',	x1=0/16, y1=5/16, x2=1, y2=6/16 )
    )
    # QR code at most 5/8 the width, to retain sufficient space for large Ethereum encrypted JSON
    # wallet private keys
    private_qr			= private.add_region_proportional(
        Image( 'private-qr-bg',	x1=0/16, y1=6/16, x2=1, y2=15/16, priority=prio_contrast )
    ).square( maximum=private.w * 5 / 8, justify='BL' )
    private_qr.add_region(
        Image( 'private-qr' )
    )
    private.add_region_proportional(
        Text( 'private-qr-b',	x1=0/16, y1=15/16, x2=1, y2=16/16 )
    )

    # Hint above; but same computed width as private_qr
    private_h_t			= private.add_region_proportional(
        Text( 'private-hint-t',	x1=0/16, y1=0/16, x2=1, y2=1/16 )
    )
    private_h_t.x2		= private_qr.x2
    private_h_bg		= private.add_region_proportional(
        Image( 'private-hint-bg',x1=0/16, y1=1/16, x2=1, y2=4/16, priority=prio_contrast )
    )
    private_h_bg.x2		= private_qr.x2
    private_h			= private.add_region_proportional(
        Text( 'private-hint',	x1=0/16, y1=1/16, x2=1, y2=3/16 )
    )
    private_h.x2		= private_qr.x2

    # We'll use the right side of the private region, each line rotated 90 degrees down and right.
    # So, we need the upper-left corner of the private-bg anchored at the upper-right corner of
    # private.
    private_length		= private.h                     # line length is y-height
    private_fontsize		= 6.5				# points == 1/72 inch
    private_height		= private.x2 - private_qr.x2 - .05
    private_lineheight		= private_fontsize / PT_IN / Text.SIZE_RATIO * .9  # in.

    private.add_region(
        Image(
            'private-bg',
            x1		= private.x2,
            y1		= private.y1,
            x2		= private.x2 + private_length,
            y2		= private.y1 + private_height,
            rotate	= -90,
            priority	= prio_contrast,
        )
    )
    # Now, add private key lines down the edge from right to left, rotating each into place
    for ln in range( int( private_height // private_lineheight )):
        private.add_region(
            Text(
                f"private-{ln}",
                font	= 'mono',
                x1	= private.x2 - private_lineheight * ln,
                y1	= private.y1,
                x2	= private.x2 + private_length,
                y2	= private.y1 + private_lineheight,
                size	= private_fontsize,
                rotate	= -90,
            )
        )

    return wallet


def page_dimensions(
    pdf: FPDF,
    page_margin_mm: Optional[float] = None,   # eg. 1/4" in mm.
) -> Type[Coordinate]:
    """Compute the working page dimensions.  FPDF().epw/.eph incl. orienation, excl. margins.

    """
    if page_margin_mm is None:
        page_margin_mm		= PAGE_MARGIN * MM_IN
    return Coordinate(
        x	= pdf.epw - page_margin_mm * 2,
        y	= pdf.eph - page_margin_mm * 2
    )


def layout_components(
    pdf: FPDF,
    comp_dim: Coordinate,
    page_margin_mm: Optional[float] = None,     # eg. 1/4" in mm.
    bleed: Optional[float]	= None,	        # eg. 5%
) -> Tuple[int, Callable[[int],Tuple[int, Type[Coordinate]]]]:
    """Compute the number of components per pdf page, and a function returning the page # and component
    (x,y) for the num'th component.

    """
    if page_margin_mm is None:
        page_margin_mm		= PAGE_MARGIN * MM_IN
    # Allow 5% bleed over into page margins (to allow for slight error in paper vs. comp sizes)
    if bleed is None:
        bleed			= 5/100
    # FPDF().epw/.eph is *without* page margins, but *with* re-orienting for portrait/landscape
    comp_cols			= int(( pdf.epw - page_margin_mm * 2 * min( 1, 1 - bleed )) // comp_dim.x )
    comp_rows			= int(( pdf.eph - page_margin_mm * 2 * min( 1, 1 - bleed )) // comp_dim.y )
    comps_pp			= comp_rows * comp_cols

    # Compute actual page margins to center cards.
    page_margin_tb		= ( pdf.eph - comp_rows * comp_dim.y ) / 2
    page_margin_lr		= ( pdf.epw - comp_cols * comp_dim.x ) / 2

    def page_xy( num: int ) -> Tuple[int, Coordinate]:
        """Returns the page, and the coordinates within that page of the num'th component"""
        page,nth		= divmod( num, comps_pp )
        page_rows,page_cols	= divmod( nth, comp_cols )
        offsetx			= page_margin_lr + page_cols * comp_dim.x
        offsety			= page_margin_tb + page_rows * comp_dim.y
        log.debug( f"{ordinal(num)} {comp_dim.x:7.5f}mm x {comp_dim.y:7.5f}mm component on page {page}, offset {offsetx:7.5f}mm x {offsety:7.5f}mm" )
        return (page, Coordinate( x=offsetx, y=offsety ))

    return (comps_pp, page_xy)




#
# Python-slip39 -- Ethereum SLIP-39 Account Generation and Recovery
#
# Copyright (c) 2022, Dominion Research & Development Corp.
#
# Python-slip39 is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.  It is also available under alternative (eg. Commercial) licenses, at
# your option.  See the LICENSE file at the top of the source tree.
#
# Python-slip39 is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
from __future__		import annotations

import itertools
import logging

from typing		import List, Optional, Union

from shamir_mnemonic	import combine_mnemonics
from shamir_mnemonic.shamir import RANDOM_BYTES

from mnemonic		import Mnemonic			# Requires passphrase as str
from mnemonic.mnemonic	import ConfigurationError
from ..util		import ordinal, commas
from ..defaults		import BITS_DEFAULT
from .entropy		import (  # noqa F401
    shannon_entropy, signal_entropy, analyze_entropy, scan_entropy, display_entropy
)

__author__                      = "Perry Kundert"
__email__                       = "perry@dominionrnd.com"
__copyright__                   = "Copyright (c) 2022 Dominion Research & Development Corp."
__license__                     = "Dual License: GPLv3 (or later) and Commercial (see LICENSE)"

log				= logging.getLogger( __package__ )


class Mnemonicv21( Mnemonic ):
    """When trezor/python-mmnemonic is updated, we can retire this."""
    @classmethod
    def detect_language(cls, code: str) -> str:
        """Scan the Mnemonic until the language becomes unambiguous, including as abbreviation prefixes."""
        code = cls.normalize_string(code)
        possible = set(cls(lang) for lang in cls.list_languages())
        words = set(code.split())
        for word in words:
            # possible languages have candidate(s) starting with the word/prefix
            possible = set(p for p in possible if any(c.startswith( word ) for c in p.wordlist))
            if not possible:
                raise ConfigurationError(f"Language unrecognized for {word!r}")
        if len(possible) == 1:
            return possible.pop().language
        # Multiple languages match: A prefix in many, but an exact match in one determines language.
        complete = set()
        for word in words:
            exact = set(p for p in possible if word in p.wordlist)
            if len(exact) == 1:
               complete.update(exact)
        if len(complete) == 1:
            return complete.pop().language
        raise ConfigurationError(
            f"Language ambiguous between {', '.join( p.language for p in possible)}"
        )


def recover(
    mnemonics: List[str],
    passphrase: Optional[Union[str,bytes]] = None,
    using_bip39: Optional[bool]	= None,  # If a BIP-39 "backup" (default: Falsey)
    as_entropy: Optional[bool]  = None,  # .. and recover original Entropy (not 512-bit Seed)
    language: Optional[str]	= None,  # ... provide language if not default 'english'
) -> bytes:
    """Recover a master secret Seed Entropy from the supplied SLIP-39 mnemonics.  We cannot know what
    subset of these mnemonics is required and/or valid, so we need to iterate over all subset
    combinations on failure; this allows us to recover from 1 (or more) incorrectly recovered
    SLIP-39 Mnemonics, using any others available.

    We'll try to find one of the smallest subsets that satisfies the SLIP-39 recovery.  The
    resultant secret Entropy is returned as the Seed, with (not widely used) SLIP-39 decryption with
    the given passphrase.  We handle either str/bytes passphrase, and will en/decode as UTF-8 as
    necessary.

    WARNING: SLIP-39 passphrase encryption is not Trezor "Model T" compatible, and is not widely
    used; if you want to hide a wallet, use the Trezor "Hidden wallet" feature instead, where the
    passphrase for each hidden wallet is entered on the device (leave this passphrase blank).

    Optionally, if using_bip39 then generate the Seed from the recovered Seed Entropy, using BIP-39
    Seed generation.  This is the ideal way to back up an existing BIP-39 Mnemonic, I believe.  It
    results in a 20- or 33-word SLIP-39 Mnemonics, retains the BIP-39 passphrase(s) for securing the
    resultant Cryptocurrency wallet(s), and is compatible with all BIP-39 hardware wallets.  Once
    you are comfortable that you can always recover your BIP-39 Mnemonic from your SLIP-39 Mnemomnic
    Cards, you are free to destroy your original insecure and unreliable BIP-39 Mnemonic backup(s).

    """
    if passphrase is None:
        passphrase		= ""
    secret			= None
    try:
        # python-shamir-mnemonic requires passphrase as bytes (not str)
        passphrase_slip39	= b"" if using_bip39 else (
            passphrase if isinstance( passphrase, bytes ) else passphrase.encode( 'UTF-8' )
        )
        combo			= range( len( mnemonics ))
        secret			= combine_mnemonics(
            mnemonics,
            passphrase	= passphrase_slip39,
        )
    except Exception as exc:
        # Try a subset of the supplied mnemonics, to silently reject any invalid mnemonic phrases supplied
        for length in range( len( mnemonics )):
            for combo in itertools.combinations( range( len( mnemonics )), length ):
                trial		= list( mnemonics[i] for i in combo )
                try:
                    secret	= combine_mnemonics(
                        trial,
                        passphrase	= passphrase_slip39,
                    )
                    break
                except Exception:
                    pass
            if secret:
                break
        if not secret:
            # No recovery; raise the Exception produced by original attempt w/ all mnemonics
            raise exc
    log.info(
        f"Recovered {len(secret)*8}-bit SLIP-39 Seed Entropy with {len(combo)}"
        f" ({'all' if len(combo) == len(mnemonics) else ', '.join( ordinal(i+1) for i in combo)})"
        f" of {len(mnemonics)} supplied mnemonics" + (
            f"; Seed decoded from SLIP-39 (w/ no passphrase) and generated using BIP-39 Mnemonic representation w/ {'a' if passphrase else 'no'} passphrase"
            if using_bip39 else
            f"; Seed decoded from SLIP-39 Mnemonics w/ {'a' if passphrase else 'no'} passphrase"
        )
    )
    if using_bip39:
        # python-mnemonic's Mnemonic requires passphrase as str (not bytes).  This is all a NO-OP,
        # if using_bip39 is Truthy and as_entropy is Truthy, but no harm...  It checks that no
        # passphrase has been supplied in that case, as a side-effect.
        passphrase_bip39	= passphrase if isinstance( passphrase, str ) else passphrase.decode( 'UTF-8' )
        # This SLIP-39 was a "backup" of a BIP-39 Mnemonic, in a 'language' (default: "english").
        secret			= recover_bip39(
            mnemonic	= produce_bip39( entropy=secret, language=language ),
            passphrase	= passphrase_bip39,
            as_entropy	= as_entropy,
            language	= language,
        )
    return secret


def recover_bip39(
    mnemonic: str,
    passphrase: Optional[Union[str,bytes]] = None,
    as_entropy: Optional[bool]	= None,   # Recover original 128- or 256-bit Entropy (not 512-bit Seed)
    language: Optional[str]	= None,   # If desired, provide language (eg. if only prefixes are provided)
) -> bytes:
    """Recover the 512-bit BIP-39 generated seed (or the original 128- or 256-bit Seed Entropy, if
    as_entropy is True) from a single BIP-39 Mnemonic Phrase, detecting the language.  Optionally
    provide a UTF-8 string or encoded passphrase (defaults to not as_entropy, if so).

    Normalizes and validates the BIP-39 Mnemonic Phrase (which is often recovered as user input):
    - Removes excess whitespace and down-cases
    - Detects language if not provided
    - Expands unambiguous mnemonic prefixes (eg. 'ae' --> 'aerobic', 'acti' --> 'action')
    - Checks that the BIP-39 Phrase check bits are valid

    Since this would normally be used to begin deriving HD wallets, the default is the hashed,
    passphrase-decrypted seed.

    """
    if as_entropy is None:
        as_entropy		= False  # Default: recover the 512-bit derivation seed
    assert not ( bool( passphrase ) and bool( as_entropy )), \
        "When recovering original BIP-39 entropy, no passphrase may be specified"
    if passphrase is None:
        passphrase		= ""

    # Polish up the supplied mnemonic, by eliminating extra spaces, leading/trailing newline(s); Mnemonic is fragile...
    mnemonic_stripped		= ' '.join( w.lower() for w in mnemonic.strip().split( ' ' ) if w )
    if mnemonic_stripped != mnemonic:
        log.info( "BIP-39 Mnemonic Phrase stripped of unnecessary whitespace" )
    # Unfortunately, Mnemonic.detect_language was unreliable; only checked the first word and
    # english/french has ambiguous words (fixed in python-mnemonic versions >=0.20).  Mnemonic must
    # be able to unambiguously detect language with the first few un-expanded mnemonics.
    if not language:
        language		= Mnemonicv21.detect_language( mnemonic_stripped )
        log.info( f"BIP-39 Language detected: {language}" )
    m				= Mnemonic( language )
    mnemonic_expanded		= m.expand( mnemonic_stripped )
    if mnemonic_expanded != mnemonic_stripped:
        log.info( "BIP-39 Mnemonic Phrase prefixes expanded" )
    if not m.check( mnemonic_expanded ):
        unrecognized		= [ w for w in mnemonic_expanded.split() if w not in m.wordlist ]
        raise ValueError( f"BIP-39 Mnemonic check fails; {len( unrecognized )} unrecognized {m.language} words {commas( unrecognized )}" )
    if as_entropy or as_entropy is None:
        # If we want to "backup" a BIP-39 Mnemonic Phrase, we want the original entropy, NOT the derived seed!
        secret			= m.to_entropy( mnemonic_expanded )
        log.info( f"Recovered {len(secret)*8}-bit BIP-39 entropy from {language} mnemonic (no passphrase supported)" )
    else:
        # python-mnemonic's Mnemonic requires passphrase as str (not bytes)
        passphrase_bip39	= passphrase if isinstance( passphrase, str ) else passphrase.decode( 'UTF-8' )
        # Only a fully validated BIP-39 Mhemonic Phrase must ever be used here!  No checking is done
        # by Mnemonic.to_seed of either the Mnemonic Phrase or passphrase (except UTF-8 encoding).
        secret			= Mnemonic.to_seed( mnemonic_expanded, passphrase = passphrase_bip39 )
        log.info( f"Recovered {len(secret)*8}-bit BIP-39 secret from {language} mnemonic{' (and passphrase)' if passphrase_bip39 else ''}" )
    return bytes( secret )  # bytearray --> bytes


def produce_bip39(
    entropy: Optional[bytes],
    strength: Optional[int]	= None,
    language: Optional[str]	= None,
) -> str:
    """Produce a BIP-38 Mnemonic from the provided entropy (or generated, default 128 bits).

    We ensure we always use the same secure entropy source from shamir_mnemonic, to allow the user
    of slip39 to monkey-patch it in one place for testing or to improve the entropy generation.
    """
    if not entropy:
        if not strength:
            strength		= BITS_DEFAULT
        entropy			= RANDOM_BYTES( strength // 8 )
    return Mnemonic( language or "english" ).to_mnemonic( entropy )



# -*- mode: python ; coding: utf-8 -*-
import json
import pytest

try:
    import eth_account
except ImportError:
    eth_account			= None
try:
    from Crypto.Cipher	import AES
    from Crypto.Protocol.KDF import scrypt
except ImportError:
    AES				= None
    scrypt			= None

import shamir_mnemonic

from .			import account, create, addresses, addressgroups, accountgroups, Account
from .recovery		import recover

from .dependency_test	import substitute, nonrandom_bytes, SEED_XMAS, SEED_ONES, SEED_ZERO

BIP39_ABANDON			= "abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon about"
BIP39_ZOO			= 'zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo wrong'

def test_account_smoke():
    acct			= account( SEED_XMAS )
    assert acct.address == '0x336cBeAB83aCCdb2541e43D514B62DC6C53675f4'
    assert acct.path == "m/44'/60'/0'/0/0"
    assert acct.key == '178870009416174c9697777b1d94229504e83f25b1605e7bb132aa5b88da64b6'

    acct			= account( SEED_XMAS, path="m/44'/60'/0'/0/1" )
    assert acct.address == '0x3b774e485fC818F0f377FBA657dfbF92B46f8504'
    assert acct.path == "m/44'/60'/0'/0/1"
    assert acct.pubkey == '03cbcf791b37011feab1c5d797cd76a3fa0f12ee5582adbe5aa4d8172a7bbaba5b'
    assert acct.key == acct.prvkey == 'bf299fe7a7d948fdb98474557bbee73395e01f3a6a73638d45d345f9adb451fb'
    assert acct.xpubkey == 'xpub6FVwqKQUrDre2ZPtAvcqR7GYW4662JTM7R1FGuwGAH3b1TjntLCbMa3HY7C1BR4ifXEtwfX63a69FEAcCSCgrgQZNd3WYgvKhAghRNucEc6'
    assert acct.xprvkey == 'xprvA2WbRosb1rJLp5KR4u5q3yKox2FbcqjVkC5eUXXebwWc8fQeLntLomiogp7vVMZcGfB4vaKeWRJ6eQmHSPRk6W7AJRNx2TT3Ai825JwH9kG'
    # And ensure we can round-trip the xprvkey (the x{pub/prv}keys encode other info, so won't be the same)
    acct			= account( acct.xprvkey, path="m/" )
    assert acct.path == 'm/'
    assert acct.address == '0x3b774e485fC818F0f377FBA657dfbF92B46f8504'
    assert acct.pubkey == '03cbcf791b37011feab1c5d797cd76a3fa0f12ee5582adbe5aa4d8172a7bbaba5b'
    assert acct.key == acct.prvkey == 'bf299fe7a7d948fdb98474557bbee73395e01f3a6a73638d45d345f9adb451fb'

    acct			= account( SEED_XMAS, crypto='Bitcoin' )
    assert acct.address == 'bc1qz6kp20ukkyx8c5t4nwac6g8hsdc5tdkxhektrt'
    assert acct.path == "m/84'/0'/0'/0/0"

    acct			= account( SEED_XMAS, crypto='Bitcoin', format='Legacy' )
    assert acct.address == '19FQ983heQEBXmopVNyJKf93XG7pN7sNFa'
    assert acct.path == "m/44'/0'/0'/0/0"
    assert acct.pubkey == '02d0bca9d976ad7303d1c0c3fd6ad0cb6bb78077d2ff158c16ac21bed763fb49a8'

    acct			= account( SEED_XMAS, crypto='Bitcoin', format='SegWit' )
    assert acct.path == "m/49'/0'/0'/0/0"
    assert acct.address == '3KbLyVYmzoDXtXMMemWV2JvvhWuWZzP5Sa'

    # And, confirm that we retrieve the same Bech32 address for the all-ones seed,
    # as on a real Trezor "Model T".
    acct			= account( SEED_ONES, crypto='Bitcoin' )
    assert acct.path == "m/84'/0'/0'/0/0"
    assert acct.address == 'bc1q9yscq3l2yfxlvnlk3cszpqefparrv7tk24u6pl'
    assert acct.pubkey == '038f7fa5776f5359eb861994bee043f0b16a5ca24b66eb38696a7325d3e1717e72'
    assert acct.prvkey == acct.key == '80d5082773a4d2a07ee667a772ca13a120a5fc9d61bcf5a32f9e7ccf731bc0e6'
    assert acct.xpubkey == 'zpub6uMZYEpdewNa98z7Hge3R4GzeayoXCmtPUzFV7DVa4cc36k2Xh7oEDvs6baStXLxT8VtXkBZ56yfuk4D5JvM43nbB7EpdkmJC75ScEZm2QK'
    assert acct.xprvkey == 'zprvAgND8jHjpZpGveueBf733vLG6Z9K7k432G4egiot1j5dAJQsz9oYgRcPFJVUdpLe3tHnabyRmmuGY871GdTv8tkotCwyn6Ec5bZbb8RjtHF'

    acct			= account( SEED_XMAS, crypto='Litecoin' )
    assert acct.address == 'ltc1qfjepkelqd3jx4e73s7p79lls6kqvvmak5pxy97'
    assert acct.path == "m/84'/2'/0'/0/0"

    acct			= account( SEED_XMAS, crypto='Litecoin', format='Legacy' )
    assert acct.address == 'LeyK1dbc5qKdKC9TvkygMTeoHixR3z1XG3'
    assert acct.path == "m/44'/2'/0'/0/0"

    acct			= account( SEED_XMAS, crypto='Litecoin', format='SegWit' )
    assert acct.path == "m/49'/2'/0'/0/0"
    assert acct.address == 'MK18ZPy9KR6PiCzti61s3HoNmEEGMmAEKc'

    acct			= account( SEED_XMAS, crypto='Dogecoin' )
    assert acct.address == 'DQCnF49GwQ5auL3u5c2uow62XS57RCA2r1'
    assert acct.path == "m/44'/3'/0'/0/0"

    acct			= account( SEED_ONES, crypto='Ripple' )
    assert acct.path == "m/44'/144'/0'/0/0"  # Default
    assert acct.address == 'rsXwvDVHHPrSm23gogdxJdrJg9WBvqRE9m'

    # Fake up some known Ripple pubkey --> addresses, by replacing the underlying "compressed"
    # public key function to return a fixed value.  Test values from:
    # trezor-firmware/core/tests/test_apps.ripple.address.py
    compressed_save		= acct.hdwallet.compressed
    acct.hdwallet.compressed	= lambda: 'ed9434799226374926eda3b54b1b461b4abf7237962eae18528fea67595397fa32'
    assert acct.pubkey == 'ed9434799226374926eda3b54b1b461b4abf7237962eae18528fea67595397fa32'
    assert acct.address == 'rDTXLQ7ZKZVKz33zJbHjgVShjsBnqMBhmN'
    acct.hdwallet.compressed	= lambda: '03e2b079e9b09ae8916da8f5ee40cbda9578dbe7c820553fe4d5f872eec7b1fdd4'
    assert acct.address == 'rhq549rEtUrJowuxQC2WsHNGLjAjBQdAe8'
    acct.hdwallet.compressed	= lambda: '0282ee731039929e97db6aec242002e9aa62cd62b989136df231f4bb9b8b7c7eb2'
    assert acct.address == 'rKzE5DTyF9G6z7k7j27T2xEas2eMo85kmw'
    acct.hdwallet.compressed	= compressed_save

    # Test values from a Trezor "Model T" w/ root seed 'zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo wrong' loaded.
    # The Trezor Suite UI produced the following account derivation path and public address for:
    acct			= account( BIP39_ZOO, crypto='Ripple' )
    assert acct.path == "m/44'/144'/0'/0/0"  # Default
    assert acct.address == 'rUPzi4ZwoYxi7peKCqUkzqEuSrzSRyLguV'
    assert acct.pubkey == '039d65db4964cbf2049ad49467a6b73e7fec7d6e6f8a303cfbdb99fa21c7a1d2bc'
    assert acct.prvkey == '6501276b9d7f646742feb12fd066e107af8c1e26e4ad7c2694279d44c43bdfb2'


def test_account_format():
    """Bitcoin test vectors from https://github.com/satoshilabs/slips/blob/master/slip-0132.md

    Since these test vectors use only the hardened path eg. m/44'/0'/0', and not the full HD
    derivation path, and since we also want to test the default paths used for various address
    formats, we'll need to use the accountgroups API (the regular account API doesn't include a
    means to restrict the path to hardened_defaults).

    """
    
    # Legacy Bitcoin
    acct			= account(
        master_secret	= BIP39_ABANDON,
        crypto		= 'Bitcoin',
        format		= "legacy",
    )
    assert acct.format == 'legacy'
    assert acct.hdwallet.seed() == "5eb00bbddcf069084889a8ab9155568165f5c453ccb85e70811aaed6f6da5fc19a5ac40b389cd370d086206dec8aa6c43daea6690f20ad3d8d48b2d2ce9e38e4"
    assert acct.path == "m/44'/0'/0'/0/0"
    assert acct.address == '1LqBGSKuX5yYUonjxT5qGfpUsXKYYWeabA'

    # m/44'/0'/0'
    (acct,),			= accountgroups(
        BIP39_ABANDON, cryptopaths=['Bitcoin'], format='legacy', hardened_defaults=True,
    )
    assert acct.format == 'legacy'
    assert acct.path == "m/44'/0'/0'"
    assert acct.xprvkey == "xprv9xpXFhFpqdQK3TmytPBqXtGSwS3DLjojFhTGht8gwAAii8py5X6pxeBnQ6ehJiyJ6nDjWGJfZ95WxByFXVkDxHXrqu53WCRGypk2ttuqncb"
    assert acct.xpubkey == "xpub6BosfCnifzxcFwrSzQiqu2DBVTshkCXacvNsWGYJVVhhawA7d4R5WSWGFNbi8Aw6ZRc1brxMyWMzG3DSSSSoekkudhUd9yLb6qx39T9nMdj"

    # m/44'/0'/0'/0/0
    acct.from_path( "m/0/0" )
    assert acct.path == "m/44'/0'/0'/0/0"
    assert acct.address == "1LqBGSKuX5yYUonjxT5qGfpUsXKYYWeabA"

    acct			= account( BIP39_ABANDON, crypto="Bitcoin", format='legacy' )
    assert acct.format == 'legacy'
    assert acct.path == "m/44'/0'/0'/0/0"
    assert acct.address == "1LqBGSKuX5yYUonjxT5qGfpUsXKYYWeabA"

    acct			= account(
        master_secret	= BIP39_ABANDON,
        crypto		= "BTC",
        path		= "...//",
        format		= "legacy",
    )
    assert acct.format == 'legacy'
    assert acct.hdwallet.semantic() == "p2pkh"
    assert acct.path == "m/44'/0'/0'"
    assert acct.xprvkey == "xprv9xpXFhFpqdQK3TmytPBqXtGSwS3DLjojFhTGht8gwAAii8py5X6pxeBnQ6ehJiyJ6nDjWGJfZ95WxByFXVkDxHXrqu53WCRGypk2ttuqncb"
    assert acct.xpubkey == "xpub6BosfCnifzxcFwrSzQiqu2DBVTshkCXacvNsWGYJVVhhawA7d4R5WSWGFNbi8Aw6ZRc1brxMyWMzG3DSSSSoekkudhUd9yLb6qx39T9nMdj"


    # SegWit
    acct			= account(
        master_secret	= BIP39_ABANDON,
        crypto		= "Bitcoin",
        format		= 'segwit'
    )
    assert acct.hdwallet.seed() == "5eb00bbddcf069084889a8ab9155568165f5c453ccb85e70811aaed6f6da5fc19a5ac40b389cd370d086206dec8aa6c43daea6690f20ad3d8d48b2d2ce9e38e4"
    assert acct.hdwallet.network() == "mainnet"
    assert acct.hdwallet.p2pkh_address() == "1PkaFBUcyAccDp2Xo2K8MqduVMgMB792r2"
    assert acct.hdwallet.p2sh_address() == "3LWUBjP2jcWhippRcyKUX1kkHNakUAy2Ms"
    assert acct.hdwallet.p2wpkh_address() == "bc1qlxgx0xk2lcjuyas4xua5p0ezg3kjfl6yd3h8y6"
    assert acct.hdwallet.p2wpkh_in_p2sh_address() == "37VucYSaXLCAsxYyAPfbSi9eh4iEcbShgf"
    assert acct.hdwallet.p2wsh_in_p2sh_address() == "3Kdr7CoTcx8UaGuzD7aqQxXi1dxUmBdph2"
    assert acct.format == "segwit"
    assert acct.hdwallet.semantic() == "p2wpkh_in_p2sh"
    assert acct.path == "m/49'/0'/0'/0/0"
    assert acct.address == "37VucYSaXLCAsxYyAPfbSi9eh4iEcbShgf"

    (acct,),			= accountgroups(
        BIP39_ABANDON, cryptopaths=['Bitcoin'], format='segwit', hardened_defaults=True,
    )
    assert acct.format == "segwit"
    assert acct.path == "m/49'/0'/0'"
    assert acct.xprvkey == "yprvAHwhK6RbpuS3dgCYHM5jc2ZvEKd7Bi61u9FVhYMpgMSuZS613T1xxQeKTffhrHY79hZ5PsskBjcc6C2V7DrnsMsNaGDaWev3GLRQRgV7hxF"
    assert acct.xpubkey == "ypub6Ww3ibxVfGzLrAH1PNcjyAWenMTbbAosGNB6VvmSEgytSER9azLDWCxoJwW7Ke7icmizBMXrzBx9979FfaHxHcrArf3zbeJJJUZPf663zsP"

    acct.from_path( "m/0/0" )
    assert acct.path == "m/49'/0'/0'/0/0"
    assert acct.address == "37VucYSaXLCAsxYyAPfbSi9eh4iEcbShgf"


    # Bech32
    (acct,),			= accountgroups(
        BIP39_ABANDON, cryptopaths=['Bitcoin'], format='bech32', hardened_defaults=True,
    )
    assert acct.format == "bech32"
    assert acct.hdwallet.semantic() == "p2wpkh"
    assert acct.path == "m/84'/0'/0'"
    assert acct.xprvkey == "zprvAdG4iTXWBoARxkkzNpNh8r6Qag3irQB8PzEMkAFeTRXxHpbF9z4QgEvBRmfvqWvGp42t42nvgGpNgYSJA9iefm1yYNZKEm7z6qUWCroSQnE"
    assert acct.xpubkey == "zpub6rFR7y4Q2AijBEqTUquhVz398htDFrtymD9xYYfG1m4wAcvPhXNfE3EfH1r1ADqtfSdVCToUG868RvUUkgDKf31mGDtKsAYz2oz2AGutZYs"

    # "m/84'/0'/0'/0/0"
    acct.from_path( "m/0/0" )
    assert acct.path == "m/84'/0'/0'/0/0"
    assert acct.address == "bc1qcr8te4kr609gcawutmrza0j4xv80jy8z306fyu"

    acct			= account(
        master_secret	= BIP39_ABANDON,
        crypto		= "BTC",
        path		= "...//",
        format		= "bech32",
    )
    assert acct.hdwallet.seed() == "5eb00bbddcf069084889a8ab9155568165f5c453ccb85e70811aaed6f6da5fc19a5ac40b389cd370d086206dec8aa6c43daea6690f20ad3d8d48b2d2ce9e38e4"
    assert acct.format == "bech32"
    assert acct.hdwallet.semantic() == "p2wpkh"
    assert acct.path == "m/84'/0'/0'"
    assert acct.xprvkey == "zprvAdG4iTXWBoARxkkzNpNh8r6Qag3irQB8PzEMkAFeTRXxHpbF9z4QgEvBRmfvqWvGp42t42nvgGpNgYSJA9iefm1yYNZKEm7z6qUWCroSQnE"
    assert acct.xpubkey == "zpub6rFR7y4Q2AijBEqTUquhVz398htDFrtymD9xYYfG1m4wAcvPhXNfE3EfH1r1ADqtfSdVCToUG868RvUUkgDKf31mGDtKsAYz2oz2AGutZYs"

    acct			= account(
        master_secret	= BIP39_ABANDON,
        crypto		= 'Bitcoin',
        format		= "bech32"
    )
    assert acct.format == "bech32"
    assert acct.hdwallet.seed() == "5eb00bbddcf069084889a8ab9155568165f5c453ccb85e70811aaed6f6da5fc19a5ac40b389cd370d086206dec8aa6c43daea6690f20ad3d8d48b2d2ce9e38e4"
    assert acct.path == "m/84'/0'/0'/0/0"
    assert acct.address == 'bc1qcr8te4kr609gcawutmrza0j4xv80jy8z306fyu'


def test_account_from_mnemonic():
    """Test all the ways the entropy 0xffff...ffff can be encoded and HD Wallets derived."""
    # Raw 0xffff...ffff entropy as Seed.  Not BIP-39 decoded (hashed) via mnemonic to produce Seed.
    # This is how SLIP-39 encodes and decodes entropy.  The raw HD Wallet Seed directly uses the
    # entropy, un-molested.
    acct_ones			= account( SEED_ONES, crypto='Bitcoin' )
    assert acct_ones.path == "m/84'/0'/0'/0/0"  # Default, BTC
    assert acct_ones.address == 'bc1q9yscq3l2yfxlvnlk3cszpqefparrv7tk24u6pl'

    details_ones_native_slip39	= create(
        "SLIP39 Wallet: Ones native SLIP-39",
        1,
        dict( fren = (3,5) ),
        SEED_ONES
    )
    # And the account addresses documented in the create's details are the SLIP-39 encodings (Seed
    # uses raw entropy)
    assert len(details_ones_native_slip39.accounts) == 1
    [(eth,btc)] = details_ones_native_slip39.accounts  # The default accounts created are ETH, BTC
    assert btc.address == 'bc1q9yscq3l2yfxlvnlk3cszpqefparrv7tk24u6pl'
    # But, we recover the raw entropy (as always from SLIP-39), except if using_bip39 is specified
    assert recover( details_ones_native_slip39.groups['fren'][1][:3] ) == SEED_ONES
    assert account( '\n'.join( details_ones_native_slip39.groups['fren'][1][:3] ), crypto='BTC' ).address == 'bc1q9yscq3l2yfxlvnlk3cszpqefparrv7tk24u6pl'
    assert account( '\n'.join( details_ones_native_slip39.groups['fren'][1][:3] ), crypto='BTC', using_bip39=True ).address == 'bc1qk0a9hr7wjfxeenz9nwenw9flhq0tmsf6vsgnn2'

    # Now 0xffff...ffff entropy as BIP-39 Seed.
    acct_ones_bip39		= account( BIP39_ZOO, crypto='BTC' )
    assert acct_ones_bip39.address == 'bc1qk0a9hr7wjfxeenz9nwenw9flhq0tmsf6vsgnn2'

    details_ones_using_bip39	= create(
        "SLIP39 Wallet: Ones using BIP-39",
        1,
        dict( fren = (3,5) ),
        SEED_ONES,
        using_bip39 = True,
    )
    # And the account addresses documented are those from the BIP-39 derived Seed
    assert len(details_ones_using_bip39.accounts) == 1
    [(eth,btc)] = details_ones_using_bip39.accounts  # The default accounts created are ETH, BTC
    assert btc.address == 'bc1qk0a9hr7wjfxeenz9nwenw9flhq0tmsf6vsgnn2'
    # But, we recover the raw entropy (as always from SLIP-39), except if using_bip39 is specified
    assert recover( details_ones_using_bip39.groups['fren'][1][:3] ) == SEED_ONES
    assert account( '\n'.join( details_ones_using_bip39.groups['fren'][1][:3] ), crypto='BTC' ).address == 'bc1q9yscq3l2yfxlvnlk3cszpqefparrv7tk24u6pl'
    assert account( '\n'.join( details_ones_using_bip39.groups['fren'][1][:3] ), crypto='BTC', using_bip39=True ).address == 'bc1qk0a9hr7wjfxeenz9nwenw9flhq0tmsf6vsgnn2'

    # Now for Ripple (XRP) accounts; first with native SLIP-39 Seed encoding
    acct			= account( SEED_ONES, crypto='Ripple' )
    assert acct.path == "m/44'/144'/0'/0/0"  # Default, XRP
    assert acct.address == 'rsXwvDVHHPrSm23gogdxJdrJg9WBvqRE9m'
    acct			= account( SEED_ONES, crypto='Ripple', path="../1" )
    assert acct.path == "m/44'/144'/0'/0/1"
    assert acct.address == 'rfMBrTc7VRTLUZu2K517r594Ky9qU5fQ5i'
    # Then with BIP-39 Seed encoding from same 0xffff...ffff entropy.  We'll test using the *same*
    # Account; the path should be set back to the default derivation path.
    acct.from_mnemonic( BIP39_ZOO )
    assert acct.path == "m/44'/144'/0'/0/0"  # Default
    assert acct.address == 'rUPzi4ZwoYxi7peKCqUkzqEuSrzSRyLguV'

    # And straight from BIP-39 Mnemonics
    details_zoos_using_bip39		= create(
        "SLIP39 Wallet: From zoo ... BIP-39",
        master_secret	= BIP39_ZOO,
    )
    [(eth_zoo,btc_zoo)]	= details_zoos_using_bip39.accounts
    btc_zoo.address == 'bc1qk0a9hr7wjfxeenz9nwenw9flhq0tmsf6vsgnn2'


@pytest.mark.skipif( not scrypt or not eth_account,
                     reason="pip install slip39[wallet] to support private key encryption" )
def test_account_encrypt():
    """Ensure BIP-38 and Ethereum JSON wallet encryption and recovery works."""

    acct			= account( SEED_XMAS, crypto='Bitcoin' )
    assert acct.address == 'bc1qz6kp20ukkyx8c5t4nwac6g8hsdc5tdkxhektrt'
    assert acct.crypto == 'BTC'
    assert acct.path == "m/84'/0'/0'/0/0"
    assert acct.legacy_address() == "134t1ktyF6e4fNrJR8L6nXtaTENJx9oGcF"

    bip38_encrypted		= acct.encrypted( 'password' )
    assert bip38_encrypted == '6PYKmUhfJa5m1NR2zUaeHC3wUzGDmb1seSEgQHK7PK5HaVRHQSp7N4ytVf'

    acct_reco			= Account( crypto='Bitcoin' ).from_encrypted( bip38_encrypted, 'password' )
    assert acct_reco.address == 'bc1qz6kp20ukkyx8c5t4nwac6g8hsdc5tdkxhektrt'
    assert acct.crypto == 'BTC'
    assert acct.path == "m/84'/0'/0'/0/0"  # The default; assumed...
    assert acct_reco.legacy_address() == "134t1ktyF6e4fNrJR8L6nXtaTENJx9oGcF"

    acct			= account( SEED_XMAS, crypto='Ethereum' )
    assert acct.address == '0x336cBeAB83aCCdb2541e43D514B62DC6C53675f4'
    assert acct.crypto == 'ETH'
    assert acct.path == "m/44'/60'/0'/0/0"

    json_encrypted		= acct.encrypted( 'password' )
    assert json.loads( json_encrypted ).get( 'address' ) == '336cbeab83accdb2541e43d514b62dc6c53675f4'

    acct_reco			= Account( crypto='ETH' ).from_encrypted( json_encrypted, 'password' )
    assert acct_reco.address == '0x336cBeAB83aCCdb2541e43D514B62DC6C53675f4'
    assert acct.crypto == 'ETH'

    # Some test cases from https://en.bitcoin.it/wiki/BIP_0038

    # No compression, no EC multiply.  These tests produce the correct private key hex, but the
    # address hash verification check fails.  I suspect that they were actually created with a
    # *different* passphrase...
    assert Account( crypto='BTC' ).from_encrypted(
        '6PRVWUbkzzsbcVac2qwfssoUJAN1Xhrg6bNk8J7Nzm5H7kxEbn2Nh2ZoGg',
        'TestingOneTwoThree',
        strict=False,
    ).key.upper() == 'CBF4B9F70470856BB4F40F80B87EDB90865997FFEE6DF315AB166D713AF433A5'

    assert Account( crypto='BTC' ).from_encrypted(
        '6PRNFFkZc2NZ6dJqFfhRoFNMR9Lnyj7dYGrzdgXXVMXcxoKTePPX1dWByq',
        'Satoshi',
        strict=False,
    ).key.upper() == '09C2686880095B1A4C249EE3AC4EEA8A014F11E6F986D0B5025AC1F39AFBD9AE'

    # This weird UTF-8 test I cannot get to pass, regardless of what format I supply the passphrase in..

    # acct_reco			= Account( crypto='BTC' ).from_encrypted(
    #     '6PRW5o9FLp4gJDDVqJQKJFTpMvdsSGJxMYHtHaQBF3ooa8mwD69bapcDQn',
    #      bytes.fromhex('cf9300f0909080f09f92a9'), # '\u03D2\u0301\u0000\U00010400\U0001F4A9'
    # )
    # assert acct_reco.legacy_address() == '16ktGzmfrurhbhi6JGqsMWf7TyqK9HNAeF'

    # No compression, no EC multiply.  These test pass without relaxing the address hash verification check.
    assert Account( crypto='BTC' ).from_encrypted(
        '6PYNKZ1EAgYgmQfmNVamxyXVWHzK5s6DGhwP4J5o44cvXdoY7sRzhtpUeo',
        'TestingOneTwoThree'
    ).key.upper() == 'CBF4B9F70470856BB4F40F80B87EDB90865997FFEE6DF315AB166D713AF433A5'

    assert Account( crypto='BTC' ).from_encrypted(
        '6PYLtMnXvfG3oJde97zRyLYFZCYizPU5T3LwgdYJz1fRhh16bU7u6PPmY7',
        'Satoshi'
    ).key.upper() == '09C2686880095B1A4C249EE3AC4EEA8A014F11E6F986D0B5025AC1F39AFBD9AE'

    # Test some addresses encrypted by our Paper Wallet PDF output procedure, using
    # the "zoo zoo ... zoo wrong" BIP-39 test Mnemonic.
    assert Account( crypto='BTC' ).from_encrypted(
        '6PYSUhj4mPTNdSvm2dxLRszieSBmzPqPQX699ECUrd69sWteFAUqmW1FLq',
        'something'
    ).address == 'bc1qk0a9hr7wjfxeenz9nwenw9flhq0tmsf6vsgnn2'

    # Ripple BIP-38 encrypted wallets.  Should round-trip via BIP-38 encryption
    acct_xrp		= Account( crypto='XRP' ).from_mnemonic( BIP39_ZOO )
    assert acct_xrp.path == "m/44'/144'/0'/0/0"  # Default
    assert acct_xrp.address == 'rUPzi4ZwoYxi7peKCqUkzqEuSrzSRyLguV'  # From Trezor "Model T" w/
    assert acct_xrp.pubkey == '039d65db4964cbf2049ad49467a6b73e7fec7d6e6f8a303cfbdb99fa21c7a1d2bc'
    acct_xrp_encrypted		= acct_xrp.encrypted( 'password' )
    assert acct_xrp_encrypted == '6PYTRxHt4sPM9i6zagBJ4pWdaefJ1FfVQwFCWQxDhVBw7fJYpYP3kMPfro'

    acct_dec			= Account( crypto='XRP' ).from_encrypted( acct_xrp_encrypted, 'password' )
    acct_dec.path == "m/44'/144'/0'/0/0"
    assert acct_dec.address == 'rUPzi4ZwoYxi7peKCqUkzqEuSrzSRyLguV'
    assert acct_dec.pubkey == '039d65db4964cbf2049ad49467a6b73e7fec7d6e6f8a303cfbdb99fa21c7a1d2bc'


@substitute( shamir_mnemonic.shamir, 'RANDOM_BYTES', nonrandom_bytes )
def test_create():
    """Standard SLIP-39 Mnemonic and account creation"""
    details_xmas		= create(
        "SLIP39 Wallet: Test",
        1,
        dict( fren = (3,5) ),
        SEED_XMAS
    )

    assert details_xmas.groups == {
        "fren": ( 3, [
            "academic acid academic acne academic academic academic academic academic academic academic academic academic academic academic academic academic carpet making building",
            "academic acid academic agree depart dance galaxy acrobat mayor disaster quick justice ordinary agency plunge should pupal emphasis security obtain",
            "academic acid academic amazing crush royal faint spit briefing craft floral negative work depend prune adapt merit romp home elevator",
            "academic acid academic arcade cargo unfold aunt spider muscle bedroom triumph theory gather dilemma building similar chemical object cinema salon",
            "academic acid academic axle crush swing purple violence teacher curly total equation clock mailman display husband tendency smug laundry disaster"
        ] ),
    }

    assert len(details_xmas.accounts) == 1
    [(eth,btc)] = details_xmas.accounts  # The default accounts created are ETH, BTC
    assert eth.address == '0x336cBeAB83aCCdb2541e43D514B62DC6C53675f4'
    assert btc.address == 'bc1qz6kp20ukkyx8c5t4nwac6g8hsdc5tdkxhektrt'

    assert recover( details_xmas.groups['fren'][1][:3] ) == SEED_XMAS

    # We can recover a slip39.Account directly from Mnemonics, too (default is native SLIP-39)
    assert account( '\n'.join( details_xmas.groups['fren'][1][:3] ), crypto='BTC' ).address == 'bc1qz6kp20ukkyx8c5t4nwac6g8hsdc5tdkxhektrt'

    # We know the native SLIP-39 and BIP-39 HD Wallet accounts for the 0xffff...ffff seed
    details_ones_native_slip39	= create(
        "SLIP39 Wallet: Native SLIP-39",
        1,
        dict( fren = (3,5) ),
        SEED_ONES
    )
    # We can recover a slip39.Account directly from Mnemonics, too (default is native SLIP-39)
    assert account( '\n'.join( details_ones_native_slip39.groups['fren'][1][:3] ), crypto='BTC' ).address == 'bc1q9yscq3l2yfxlvnlk3cszpqefparrv7tk24u6pl'

    details_ones_using_bip39	= create(
        "SLIP39 Wallet: Backup BIP-39",
        1,
        dict( fren = (3,5) ),
        SEED_ONES,
        using_bip39 = True,
    )
    # ... but recovery of "backup" of a BIP-39 via SLIP-39 is also directly supported:
    assert account( '\n'.join( details_ones_using_bip39.groups['fren'][1][:3] ), crypto='BTC', using_bip39=True ).address == 'bc1qk0a9hr7wjfxeenz9nwenw9flhq0tmsf6vsgnn2'


@substitute( shamir_mnemonic.shamir, 'RANDOM_BYTES', nonrandom_bytes )
def test_create_bip39():
    """Standard SLIP-39 Mnemonic from BIP-39 backup and account creation.

    """
    details			= create(
        "SLIP39 Wallet: Test",
        1,
        dict( fren = (3,5) ),
        SEED_ONES,
        using_bip39	= True,
        cryptopaths	= ('ETH','BTC','XRP'),
    )

    assert len(details.accounts) == 1
    [(eth,btc,xrp)] = details.accounts
    # recognizable ETH, BTC accounts generated from the 0xffff...ffff (zoo zoo ...  wrong) Seed.
    assert eth.address == '0xfc2077CA7F403cBECA41B1B0F62D91B5EA631B5E'
    assert btc.address == 'bc1qk0a9hr7wjfxeenz9nwenw9flhq0tmsf6vsgnn2'
    assert xrp.address == 'rUPzi4ZwoYxi7peKCqUkzqEuSrzSRyLguV'


def test_addresses():
    master_secret		= b'\xFF' * 16
    addrs			= list( addresses(
        master_secret	= master_secret,
        crypto		= 'ETH',
        paths		= ".../-9",
    ))
    # print( json.dumps( addrs, indent=4, default=str ))
    assert addrs == [
        (
            "ETH",
            "m/44'/60'/0'/0/0",
            "0x824b174803e688dE39aF5B3D7Cd39bE6515A19a1"
        ),
        (
            "ETH",
            "m/44'/60'/0'/0/1",
            "0x8D342083549C635C0494d3c77567860ee7456963"
        ),
        (
            "ETH",
            "m/44'/60'/0'/0/2",
            "0x52787E24965E1aBd691df77827A3CfA90f0166AA"
        ),
        (
            "ETH",
            "m/44'/60'/0'/0/3",
            "0xc2442382Ae70c77d6B6840EC6637dB2422E1D44e"
        ),
        (
            "ETH",
            "m/44'/60'/0'/0/4",
            "0x42a910D380dE132B5227e3277Cc70C3C76a884aC"
        ),
        (
            "ETH",
            "m/44'/60'/0'/0/5",
            "0x1A3db5E0422c78F43a35686f0307Da8f22344dE0"
        ),
        (
            "ETH",
            "m/44'/60'/0'/0/6",
            "0x19031c515C5d91DB7988D89AAA6F71a5825f5245"
        ),
        (
            "ETH",
            "m/44'/60'/0'/0/7",
            "0xaE693156ac600f5B0D58e5090ecf0A578c5Cc0a8"
        ),
        (
            "ETH",
            "m/44'/60'/0'/0/8",
            "0x4347541fa648BCE62543a8AbC2901E08017f6A6a"
        ),
        (
            "ETH",
            "m/44'/60'/0'/0/9",
            "0xC11235559Dd4c5224a19396C3A14526E92ebba35"
        )
    ]
    addrs			= list( addresses(
        master_secret	= master_secret,
        crypto		= 'BTC',
        paths		= '.../0/-2',
        format		= 'Bech32',
    ))
    #print( json.dumps( addrs, indent=4, default=str ))
    assert addrs == [
        (
            "BTC",
            "m/84'/0'/0'/0/0",
            "bc1q9yscq3l2yfxlvnlk3cszpqefparrv7tk24u6pl"
        ),
        (
            "BTC",
            "m/84'/0'/0'/0/1",
            "bc1qnec684yvuhfrmy3q856gydllsc54p2tx9w955c"
        ),
        (
            "BTC",
            "m/84'/0'/0'/0/2",
            "bc1q2snj0zcg23dvjpw7m9lxtu0ap0hfl5tlddq07j"
        ),
    ]


def test_addressgroups():
    master_secret		= BIP39_ZOO
    addrgrps			= list( enumerate( addressgroups(
        master_secret	= master_secret,
        cryptopaths	= [
            ('ETH',	".../-3"),
            ('BTC',	".../-3"),
            ('LTC',	".../-3"),
            ('Doge',	".../-3"),
            ('Binance',	".../-3"),
            ('Ripple',	".../-3"),
        ],
    )))
    # print( repr( addrgrps ))
    assert addrgrps == [								# Verified
        (0,(('ETH', "m/44'/60'/0'/0/0", '0xfc2077CA7F403cBECA41B1B0F62D91B5EA631B5E'),	# Ledger
            ('BTC', "m/84'/0'/0'/0/0", 'bc1qk0a9hr7wjfxeenz9nwenw9flhq0tmsf6vsgnn2'),	# Ledger
            ('LTC', "m/84'/2'/0'/0/0", 'ltc1qnreu4d88p5tvh33anptujvcvn3xmfhh43yg0am'),	# Ledger
            ('DOGE', "m/44'/3'/0'/0/0", 'DTMaJd8wqye1fymnjxZ5Cc5QkN1w4pMgXT'),		# Ledger
            ('BSC', "m/44'/60'/0'/0/0", '0xfc2077CA7F403cBECA41B1B0F62D91B5EA631B5E'),	# Ledger
            ('XRP', "m/44'/144'/0'/0/0", 'rUPzi4ZwoYxi7peKCqUkzqEuSrzSRyLguV'))),	# Ledger 
        (1, (('ETH', "m/44'/60'/0'/0/1", '0xd1a7451beB6FE0326b4B78e3909310880B781d66'),
             ('BTC', "m/84'/0'/0'/0/1", 'bc1qkd33yck74lg0kaq4tdcmu3hk4yruhjayxpe9ug'),
             ('LTC', "m/84'/2'/0'/0/1", 'ltc1qm4yc8vgxyv0xeu8p4vtq2wls245y2ueqpfrp4d'),
             ('DOGE', "m/44'/3'/0'/0/1", 'DGkL2LD5FfccAaKtx8G7TST5iZwrNkecTY'),
             ('BSC', "m/44'/60'/0'/0/1", '0xd1a7451beB6FE0326b4B78e3909310880B781d66'),
             ('XRP', "m/44'/144'/0'/0/1", 'ravkJwvQBuW4P5TG1qK5WDAgBxbPhdyPh1'))),
        (2, (('ETH', "m/44'/60'/0'/0/2", '0x578270B5E5B53336baC354756b763b309eCA90Ef'),
             ('BTC', "m/84'/0'/0'/0/2", 'bc1qvr7e5aytd0hpmtaz2d443k364hprvqpm3lxr8w'),
             ('LTC', "m/84'/2'/0'/0/2", 'ltc1qstkxz076qdyg0r08eszf0rrxsmfcgj62lkqaj2'),
             ('DOGE', "m/44'/3'/0'/0/2", 'DQa3SpFZH3fFpEFAJHTXZjam4hWiv9muJX'),
             ('BSC', "m/44'/60'/0'/0/2", '0x578270B5E5B53336baC354756b763b309eCA90Ef'),
             ('XRP', "m/44'/144'/0'/0/2", 'rpzdHCsqVLppnUAUvgYDd6ADZFeKE6QoHR'))),
        (3, (('ETH', "m/44'/60'/0'/0/3", '0x909f59835A5a120EafE1c60742485b7ff0e305da'),
             ('BTC', "m/84'/0'/0'/0/3", 'bc1q6t9vhestkcfgw4nutnm8y2z49n30uhc0kyjl0d'),
             ('LTC', "m/84'/2'/0'/0/3", 'ltc1qts5sde8st3x6qt2t0xhtf9uactg7nnztamuehk'),
             ('DOGE', "m/44'/3'/0'/0/3", 'DTW5tqLwspMY3NpW3RrgMfjWs5gnpXtfwe'),
             ('BSC', "m/44'/60'/0'/0/3", '0x909f59835A5a120EafE1c60742485b7ff0e305da'),
             ('XRP', "m/44'/144'/0'/0/3", 'r9czvGVozoKTAnP1G17RG9DfWK272ZExvX')))
    ]


def test_accountgroups():
    master_secret		= b'\xFF' * 16
    acctgrps			= list( accountgroups(
        master_secret	= master_secret,
        cryptopaths	= [
            ('ETH', ".../-3"),
            ('BTC', ".../-3"),
        ],
    ))
    # print( json.dumps( acctgrps, default=repr ))
    assert len(acctgrps) == 4




#
# Python-slip39 -- Ethereum SLIP-39 Account Generation and Recovery
#
# Copyright (c) 2022, Dominion Research & Development Corp.
#
# Python-slip39 is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.  It is also available under alternative (eg. Commercial) licenses, at
# your option.  See the LICENSE file at the top of the source tree.
#
# Python-slip39 is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#

from __future__		import annotations

import json
import logging
import os
import traceback
import time

from dataclasses	import dataclass
from datetime		import datetime
from enum		import Enum
from fractions		import Fraction
from hashlib		import sha256
from pathlib		import Path
from typing		import Dict, Optional, Union, Tuple

import requests
import eth_account
import eth_abi
import solcx

from rlp		import encode as rlp_encode
from web3		import Web3
from web3.middleware	import construct_sign_and_send_raw_middleware

from ..api		import Account
from ..util		import memoize, retry, commas, into_bytes, timer
from ..defaults		import (
    ETHERSCAN_MEMO_MAXAGE, ETHERSCAN_MEMO_MAXSIZE,
    TOKPRICES_MEMO_MAXAGE, TOKPRICES_MEMO_MAXSIZE,
    INVOICE_PROXIES,
)


__author__                      = "Perry Kundert"
__email__                       = "perry@dominionrnd.com"
__copyright__                   = "Copyright (c) 2022 Dominion Research & Development Corp."
__license__                     = "Dual License: GPLv3 (or later) and Commercial (see LICENSE)"

log				= logging.getLogger( 'ethereum' )


def contract_address(
    address,			# Address that is constructing the contract
    salt		= None,
    creation		= None,
    creation_hash	= None,
    nonce		= None,		# traditional CREATE used address/nonce
):
    """Deduces the Contract Address that will result from a creator's 'address' and a
    transaction/CREATE (given a 'nonce'), or CREATE2 (given a 'salt' and the contract's 'creation'
    bytecode).

    """
    b_address			= into_bytes( address )
    assert isinstance( b_address, bytes ) and len( b_address ) == 20, \
        f"Expected 20-byte adddress, got {b_address!r}"

    if nonce is not None and salt is None and creation is None and creation_hash is None:
        # A CREATE (traditional, or transaction-based) contract creation; only address and nonce
        assert isinstance( nonce, int ), \
            f"The nonce for CREATE must be an integer, not {nonce!r}"
        b_result		= Web3.keccak( rlp_encode([ b_address, nonce ]) )
    else:
        # A CREATE2 (deterministic) contract creation
        assert salt is not None and ( creation is not None or creation_hash is not None ) and nonce is None, \
            "Need salt and creation bytecode/hash for CREATE2"
        b_pre			= into_bytes( '0xff' )
        if isinstance( salt, int ):
            b_salt		= eth_abi.encode( [ 'uint256' ], [ salt ] )
        else:
            b_salt		= into_bytes( salt )
        assert len( b_salt ) == 32, \
            f"Expected 32-byte salt, got {len(b_salt)} bytes"
        if creation_hash:
            b_creation_hash	= into_bytes( creation_hash )
        else:
            b_creation_hash	= Web3.keccak( into_bytes( creation ))
        b_result		= Web3.keccak( b_pre + b_address + b_salt + b_creation_hash )

    result_address		= Web3.to_checksum_address( b_result[12:].hex() )

    return result_address


class Chain( Enum ):
    Nothing		= 0
    Ethereum		= 1
    Goerli		= 2


#
# Etherscan API, for accessing Gas Oracle and wallet data
#
etherscan_urls			= dict(
    Ethereum	= 'https://api.etherscan.io/api',
    Goerli	= 'https://api-goerli.etherscan.io/api',
)

#
# Alchemy API, for accessing the Ethereum blockchain w/o running a local instance
#
#     They provide a free "testing" API token, with a very low 50 "Compute Units/Second" limit, for
# testing their API: https://docs.alchemy.com/reference/throughput.  This should be sufficient for
# accessing "free" (view) APIs in contracts.  However, for any advanced usage (ie. to deploy
# contracts), you'll probably need an official ALCHEMY_API_TOKEN
#
alchemy_urls			= dict(
    Ethereum	= 'eth-mainnet.g.alchemy.com/v2',
    Goerli	= 'eth-goerli.g.alchemy.com/v2',
)

alchemy_api_testing		= dict(
    Goerli	= 'AxnmGEYn7VDkC4KqfNSFbSW9pHFR7PDO',
    Ethereum	= 'J038e3gaccJC6Ue0BrvmpjzxsdfGly9n',
)


def alchemy_url( chain, protocol='wss' ):
    """Return our Alchemy API URL, including our API token, from the ALCHEMY_API_TOKEN environment
    variable.

    If none specified, we'll default to their "Testing" API token, which should be adequate for
    low-rate "free" (view) contract APIs.  However, for deploying contracts, either an official
    (free) Alchemy API token will be required, or a local Ethereum node.

    """
    assert protocol in ( 'wss', 'https' ), \
        "Must specify either Websocket over SSL ('wss') or HTTP over SSL ('https')"
    api_token			= os.getenv( 'ALCHEMY_API_TOKEN' )
    if not alchemy_url.api_token or api_token != alchemy_url.api_token:
        # Alchemy API token not yet discovered, or has changed
        if api_token:
            alchemy_url.api_token	= api_token
            log.info( f"Using supplied Alchemy {chain} API token: {alchemy_url.api_token:.5}..." )
        elif not alchemy_url.api_token:
            alchemy_url.api_token	= alchemy_api_testing[chain.name]
            log.warning( f"Using \"Testing\" Alchemy {chain} API token: {alchemy_url.api_token};"
                         " obtain an official API key: https://docs.alchemy.com/reference/api-overview" )
    return f"{protocol}://{alchemy_urls[chain.name]}/{alchemy_url.api_token}"
alchemy_url.api_token		= None   # noqa E305


@memoize( maxage=ETHERSCAN_MEMO_MAXAGE, maxsize=ETHERSCAN_MEMO_MAXSIZE, log_at=logging.INFO )
def etherscan( chain, params, headers=None, apikey=None, timeout=None, verify=True ):
    """Queries etherscan.io, optionally w/ your apikey.  Must specify name of Ethereum blockchain to
    use.  The params must be a hashable sequence (tuple of tuples) usable to construct a dict, since
    memoize only caches based on args, and all args must be hashable.

    Raises exception on timeout, absence of successful response, absence of 'result' in response.
    Does no other checking on the content of the response' 'result'.

    Without an API key, request rates are severely restricted to ~1/5s, eg.:
        {
            "status": "1",
            "message": "OK-Missing/Invalid API Key, rate limit of 1/5sec applied",
            "result": {
                "ethbtc": "0.07217",
                "ethbtc_timestamp": "1671982262",
                "ethusd": "1214.6",
                "ethusd_timestamp": "1671982258"
            }
        }

    or:

        {
            "status": "0",
            "message": "NOTOK",
            "result": "Max rate limit reached, please use API Key for higher rate limit"
        }


    """
    assert chain.name in etherscan_urls, \
        f"No API service URL specified for {chain}"
    url				= etherscan_urls[chain.name]
    timeout			= timeout or 5.0
    headers			= headers or {
        'Content-Type':  'application/x-javascript',
    }
    params			= dict( params )
    if apikey is None:
        apikey			= os.getenv( 'ETHERSCAN_API_TOKEN' )
    if apikey and apikey.strip():  # May remain None, or be empty
        params.setdefault( 'apikey', apikey.strip() )

    # A successful request is a 200 OK, with a JSON-encoded result dict/ w a status: "1".  We do not
    # want to return any non-Exception response excepts successes, because these are Memoized.
    log.debug( "Querying {} w/ {}".format( url, params ))
    try:
        response		= requests.get(
            url,
            params	= params,
            headers	= headers,
            timeout	= timeout,
            verify	= verify,
        )
        assert response.status_code == 200, \
            "Failed to query {} for {}: {}".format( chain, params, response.text )
        response_json	= response.json()
        assert hasattr( response_json, 'keys' ) and {'status', 'result'} <= set( response_json.keys() ) and int( response_json['status'] ), \
            "Query {} for {} yielded invalid response: {}".format( chain, params, response.text )
    except Exception as exc:
        log.info( f"Query failed w/ Exception: {exc}" )
        raise

    # OK, got a valid, successful response we are prepared to Memoize!
    log.info( "Querying {} w/ {}: {}".format(
        url, params,
        json.dumps( response_json, indent=4 ) if log.isEnabledFor( logging.DEBUG ) else response.text
    ))
    return response_json['result']


@retry( tries=5, delay=3, backoff=1.5, log_at=logging.INFO, exc_at=logging.WARNING, default_cls=dict )
def gasoracle( chain=None, **kwds ):
    """Return (possibly cached) Gas Oracle values from etherscan.io, or empty dict, allowing retries w/
    up to 3*1.5^5 seconds (22s) exponential backoff.

    """
    return etherscan(
        chain or Chain.Ethereum,
        (
            ('module', 'gastracker'),
            ('action', 'gasoracle'),
        ),
        **kwds,
    )


@retry( tries=5, delay=3, backoff=1.5, log_at=logging.INFO, exc_at=logging.WARNING, default_cls=dict )
def ethprice( chain=None, **kwds ):
    """Return (possibly cached) Ethereum price in $USD from etherscan.io, or empty dict (performs exponential
    backoff of 3*1.5^5 seconds (22s) on Exceptions.)

    """
    return etherscan(
        chain or Chain.Ethereum,
        (
            ('module', 'stats'),
            ('action', 'ethprice'),
        ),
        **kwds,
    )


@retry( tries=5, delay=3, backoff=1.5, log_at=logging.INFO, exc_at=logging.WARNING, default_cls=dict )
def erc20tx( chain=None, address=None, token=None, **kwds ):
    """Return (possibly cached) ERC-20 transactions from etherscan.io, or empty dict (performs exponential
    backoff of 3*1.5^5 seconds (22s) on Exceptions.)

    Caches based on account address and (optionally) a specific token address.  Otherwise, returns
    all (known) ERC-20 token operations on the specified account.

    TODO: There may be unknown ERC-20 tokens that this won't find, unless contractAddress specified.

    """
    assert address is not None, \
        "Require an Ethereum account 'address' to query for ERC-20 transactions"
    if token is None:
        params			= (
            ('module', 'account'),
            ('action', 'tokentx'),
            ('address', address),
        )
    else:
        params			= (
            ('module', 'account'),
            ('action', 'tokentx'),
            ('address', address),
            ('contractAddress', token),
        )
    return etherscan(
        chain or Chain.Ethereum,
        params,
        **kwds,
    )


@retry( tries=5, delay=3, backoff=1.5, log_at=logging.INFO, exc_at=logging.WARNING, default_cls=dict )
def etherbalance( chain=None, address=None, **kwds ):
    """Return (possibly cached) ETH balance from etherscan.io, or empty dict (performs exponential
    backoff of 3^5 seconds (4 min.) on Exceptions.)

    Caches based on account address.

    """
    assert address is not None, \
        "Require an Ethereum account 'address' to query for Ethereum balance"
    params			= (
        ('module', 'account'),
        ('action', 'balance'),
        ('address', address),
    )
    return etherscan(
        chain or Chain.Ethereum,
        params,
        **kwds,
    )


@retry( tries=5, delay=3, backoff=1.5, log_at=logging.INFO, exc_at=logging.WARNING, default_cls=dict )
def ethertx( chain=None, address=None, **kwds ):
    """Return (possibly cached) ETH normal transactions from etherscan.io, or empty dict (performs exponential
    backoff of 3^5 seconds (4 min.) on Exceptions.)

    Caches based on account address.

    Includes contract invocations, incoming/outgoing value transactions, etc., including those that
    have "isError": "1".

    """
    assert address is not None, \
        "Require an Ethereum account 'address' to query for Ethereum transactions"
    params			= (
        ('module', 'account'),
        ('action', 'txlist'),
        ('address', address),
    )
    return etherscan(
        chain or Chain.Ethereum,
        params,
        **kwds,
    )


class Direction( Enum ):
    Incoming		= 0b01
    Outgoing		= 0b10
    Both		= 0b11


def etherio( chain=None, address=None, direction=None, **kwds ):
    """Yields a sequence of successful Ethereum value incoming/outgoing transactions for an account.
    By default, returns only Incoming value transactions.

    Ignores non-value (eg. contract) and failed transactions.

    """
    if direction is None:
        direction		= Direction.Incoming
    for tx in ethertx( chain=chain, address=address, **kwds ):
        if tx["value"] == "0" or tx["isError"] != "0":
            continue
        # A non-error value-bearing transaction.  Ignore?
        if not ( direction == Direction.Both
                 or (( direction.value & Direction.Incoming.value ) and tx["to"  ].lower() == address.lower() )
                 or (( direction.value & Direction.Outgoing.value ) and tx["from"].lower() == address.lower() )):
            continue
        yield tx


def erc20io( chain=None, address=None, token=None, tokens=None, direction=None, **kwds ):
    """Yields a sequence of successful ERC-20 value incoming/outgoing transactions for an account.
    By default, returns only Incoming value transactions.

    Specifying a specific 'token' limits the query to only a single token.  Alternatively,
    specifying a set/list of tokens filters transactions to only those tokens.

    """
    assert not bool( token ) or not bool( tokens ) or token in tokens, \
        "Cannot specify both a 'token' and 'tokens' set that do not overlap"
    if direction is None:
        direction		= Direction.Incoming
    tokens			= set( t.lower() for t in tokens ) if tokens else set()
    for tx in erc20tx( chain=chain, address=address, token=token ):
        if tokens and tx["contractAddress"].lower() not in tokens:
            continue
        if not ( direction == Direction.Both
                 or (( direction.value & Direction.Incoming.value ) and tx["to"  ].lower() == address.lower() )
                 or (( direction.value & Direction.Outgoing.value ) and tx["from"].lower() == address.lower() )):
            continue
        yield tx


class Speed( Enum ):
    Propose		= 0
    Safe		= 1
    Fast		= 2


class GasOracle:
    GWEI_WEI			= 10 ** 9		# GWEI, in WEIs
    ETH_GWEI			= 10 ** 9		# ETH, in GWEIs
    ETH_WEI			= 10 ** 18		# ETH, in WEIs

    # Some defaults; maintain the invariant BASE_... + PRIORITY_... == GAS_...
    GAS_GWEI_DEFAULT		= 12.0
    BASEFEE_GWEI_DEFAULT	= 10.0
    ETH_USD_DEFAULT		= 1000.0

    @property
    def ETH_USD( self ):
        """Return some estimate of ETH value in USD$, for use in Gas Price calculations."""
        return self.ETH_USD_DEFAULT

    def __bool__( self ):
        """Return falsey until this GasOracle is available.  Good for waiting for Gas Oracle sources
        with API rate limits, etc, or just to use this as a default that is never ready.

        """
        return False

    def maxPriorityFeePerGas( self, spend=None, gas=None, max_factor=None ):
        """At the very least, computes EIP-1559 Gas pricing, in Wei.  If spend/gas supplied, should
        also include maxFeePerGas we're willing to spend for this transaction.

        """
        raise NotImplementedError()


class Etherscan( GasOracle ):
    """Retrieve (or supply some defaults for) Gas and Ethereum pricing and some useful constants, IF
    you supply an etherscan.io API token in the ETHERSCAN_API_TOKEN environment variable and have
    network access.  Defaults to the Ethereum chain.  Implements the GasOracle API.

    If Etherscan.UPDATED or .STATUS are falsey, this indicates that the value(s) are estimated;
    otherwise, they will return the approximate *nix timestamp (or UTC time) of the provided Gas and
    Ethereum pricing, eg:

        >>> ETH = Etherscan( "Ethereum )
        >>> log.warning( f"Ethereum price: USD${ETH.ETH_USD:7.2f} ({ETH.STATUS or 'estimated'})" )

    To simplify Ethereum transaction gas pricing, we compute a proposed maxPriorityFeePerGas value;
    the only value required for new-style Ethereum transactions -- the "tip" we'll propose to
    influence our transaction to be included.  By default, we'll use the "ProposeGasPrice"
    (moderate tip).  Optionally, you may specify a Speed.Safe (slow, but should happen sometime
    soon) or Speed.Fast (should be prioritised and happen quickly).

    NOTE: Take care to ensure that any Wei values are integer; any larger denominatinos (eg. Gwei)
    should be able to handle fractional values.
    """
    def __init__( self, chain=None, speed=None ):
        if chain and isinstance( chain, str ):
            chain,		= ( c for c in Chain if c.name.lower() == chain.lower() )
        assert isinstance( chain, (Chain, type(None)) )
        self._chain		= chain or Chain.Ethereum

        if speed and isinstance( speed, str ):
            speed,		= ( s for s in Speed if s.name.lower() == speed.lower() )
        assert isinstance( speed, (Speed, type(None)) )
        self.speed		= speed or Speed.Propose

    @classmethod
    def reset( cls ):
        """Clears all memoized data and retry back-offs, forcing a refresh on next call."""
        etherscan.reset()
        ethprice.reset()
        gasoracle.reset()

    @property
    def chain( self ):
        return self._chain

    @chain.setter
    def chain( self, chain ):
        """Change the chain, flushing any memoize/retry result caching."""
        self._chain		= chain
        self.reset()

    @property
    def ETH( self ):
        """Return Ethereum price information, or empty dict"""
        return ethprice( chain=self.chain )

    @property
    def TIMESTAMP( self ):
        """Indicates if the Ethereum price self.ETH has been queried."""
        return int( self.ETH.get( 'ethusd_timestamp', 0 ))

    @property
    def ETH_USD( self ):
        """ETH, in USD$"""
        return float( self.ETH.get( 'ethusd', self.ETH_USD_DEFAULT ))

    @property
    def GAS( self ):
        """Return Gas price oracle information, or empty dict"""
        return gasoracle( chain=self.chain )

    @property
    def LASTBLOCK( self ):
        """Ethereum Gas Fee estimate comes from this block; indicates if the self.GAS has been queried."""
        return int( self.GAS.get( 'LastBlock', 0 ))

    @property
    def GAS_GWEI( self ):
        """Ethereum aggregate (pre-EIP-1559) Gas Fee proposed, in GWEI.  This is the total Base +
        Priority fee to be paid, in Gwei per Gas, for the specified transaction Speed desired.  This
        would be the value you'd use for a traditional-style transaction with a "gasPrice" value.

        If we don't know, we'll just return a guess.  It is likely that you'll want to check if
        self.UPDATED, and do something else to compute your gas pricing -- such as ask your Ethereum
        blockchain what it suggests, via the JSON eth_maxPriorityFeePerGas API call, for example.

        """
        return float( self.GAS.get( self.speed.name+'GasPrice', self.GAS_GWEI_DEFAULT ))

    @property
    def GAS_WEI( self ):
        """Ethereum Gas Fee, in WEI"""
        return int( self.GAS_GWEI * self.GWEI_WEI )

    @property
    def GAS_USD( self ):
        """Computes the cost per Gas, in USD$.  Use this to calculate a maxFeePerGas value, if you want to
        limit the cost for a transaction to some dollar amount (for the gas limit you've specified
        for the transaction).  If it turns out that the required transaction gas fees exceed the
        maxFeePerGas you've specified, the transaction will be rejected (and won't spend your gas!)

        For example, lets say you have a transaction you know will spend (less than) 100,000 gas so
        you specify a 'gas' limit of 100000.  But, you can wait, so you don't want to spend too much
        on the transaction, say less than USD$1.50.

            >>> gas, spend = 100000, 1.50  # USD$
            >>> ETH = Etherscan( "Ethereum )
            >>> maxFeePerGas = spend / ETH.GAS_USD ETH.GAS_WEI (TODO)

        """
        return self.ETH_USD * self.GAS_GWEI / self.ETH_GWEI

    @property
    def BASEFEE_GWEI( self ):
        """Ethereum Base Fee estimate for upcoming block, in GWEI (Etherscan API returns fees in Gwei/Gas).

        """
        return float( self.GAS.get( 'suggestBaseFee', self.BASEFEE_GWEI_DEFAULT ))

    @property
    def BASEFEE_WEI( self ):
        """Ethereum Base Fee, in WEI"""
        return int( self.BASEFEE_GWEI * self.GWEI_WEI )

    @property
    def PRIORITY_GWEI( self ):
        """Ethereum Priority Fee per Gas estimate for upcoming block, in GWEI.  Compute this by taking the
        Safe/Propose/Fast total Gas Price Fee estimate, and subtracting off the estimated Base Fee.

        This is what you'd supply if you want to use maxPriorityFeePerGas for your Ethereum
        transaction (actually, since they usually want it in Wei, use self.PRIORITY_WEI).

        """
        return self.GAS_GWEI - self.BASEFEE_GWEI

    @property
    def PRIORITY_WEI( self ):
        return int( self.PRIORITY_GWEI * self.GWEI_WEI )

    @property
    def UPDATED( self ):
        """If we successfully obtained Gas and Ethereum pricing, return the timestamp of the Ethereum."""
        return self.LASTBLOCK and self.TIMESTAMP

    @property
    def STATUS( self ):
        updated			= self.UPDATED
        if updated:
            return datetime.utcfromtimestamp( updated ).ctime() + " UTC"

    def __bool__( self ):
        """The GasOracle API requires a True result when maxPriorityFeePerGas is available."""
        return bool( self.UPDATED )

    def maxFeePerGas( self, spend=None, gas=None, max_factor=None ):
        """Returns new-style EIP-1559 max gas fees allowed, in Wei.  Computes a maxFeePerGas we're
        willing to pay per Gas, for max 'gas', to keep total transaction cost below 'spend'.

        If either is not specified, we'll simply cap it at a base fee + 'max_factor' x priority fee.

        Results are expressed in integer Wei.
        """
        assert isinstance( spend, (int,float,type(None))), \
            f"Require a numeric spend amount in USD$, not {spend!r}"
        assert isinstance( gas, (int,float,type(None))), \
            f"Require a numeric gas amount, not {gas!r}"
        if not spend or not gas:
            # Default: cap at some additional factor of the priority fee.
            return dict(
                maxFeePerGas	= self.BASEFEE_WEI + self.PRIORITY_WEI * ( max_factor or 2 )
            )

        # Otherwise, we can compute what we're willing to pay per Gas, from the specified total
        # spend allowed vs. gas estimate.  The transaction will not cost more than the specified
        # spend.  Requires a reasonable estimate of gas limit vs. the actual transaction Gas cost.

        # Eg. USD$1.50 * 10^9 / USD$1,000 == 1,500,000 Gwei to use on Gas
        gwei_available			= spend
        gwei_available		       *= self.ETH_GWEI
        gwei_available		       /= self.ETH_USD
        # 1.5e6 Gwei / 21,000 Gas == 71.4 Gwei/Gas max gas fee allowance
        gwei_per_gas			= gwei_available / gas
        wei_per_gas			= int( gwei_per_gas * self.GWEI_WEI )
        return dict(
            maxFeePerGas		= wei_per_gas
        )

    def maxPriorityFeePerGas( self, spend=None, gas=None, max_factor=None ):
        """Returns new-style EIP-1559 gas Priority fee we'll likely need to pay, in Wei, to achieve the
        desired Speed (Safe, Propose or Fast).  This Priority Fee (plus the current network Base
        Fee) will be the Gas cost we'll pay, in Wei/Gas.

        If spend and gas is supplied, also includes the specific maxFeePerGas we're willing to spend.

        Results are expressed in integer Wei.
        """
        gas_price			= dict(
            maxPriorityFeePerGas	= self.PRIORITY_WEI,
        ) | self.maxFeePerGas( spend=spend, gas=gas, max_factor=max_factor )
        log.info(
            f"{self.chain}: EIP-1559 Gas Pricing at USD${self.ETH_USD:9,.2f}/ETH:"
            f" {gas_price['maxPriorityFeePerGas'] / self.GWEI_WEI:9,.2f} Priority + {self.BASEFEE_GWEI:9,.2f} Base Gwei/Gas"
            + (
                f"; for max USD${spend or gas * gas_price['maxFeePerGas'] * self.ETH_USD / self.ETH_WEI:9,.2f} per"
                f" {gas:10,}-Gas transaction: {gas_price['maxFeePerGas'] / self.GWEI_WEI:9,.2f} Gwei/Gas"
                if gas and 'maxFeePerGas' in gas_price
                else ""
            )
        )
        return gas_price

    def gasPrice( self ):
        """The traditional gasPrice interface.  Returns the total gasPrice we're willing to pay.

        """
        gas_price			= dict(
            gasPrice			= self.GAS_WEI,
        )
        log.info(
            f"{self.chain}: Traditional Gas Pricing: {gas_price['gasPrice'] / self.GWEI_WEI:9,.2f} Gwei/Gas"
        )
        return gas_price


#
# Optimize, and search for imports relative to this directory.
#
# Don't over-optimize, or construction code gets very big (and expensive to deploy)
#
solcx_options			= dict(
    optimize		= True,
    optimize_runs	= 100,
    # Any contract paths referenced from within contracts search relative to this
    base_path		= Path( __file__ ).resolve().parent
)
solc_version			= "0.8.17"


class Contract:
    """Interact with Ethereum contracts, via web3.py's web3.Web3 instance.

    Specify an ETHERSCAN_API_TOKEN to avoid the free tier's 1/5s call rate.  We always query the
    Ethereum chain via Etherscan (there is no gasoracle on it's Goerli API URL).

    Specify an ALCHEMY_API_TOKEN and use Alchemy's Goerli or Ethereum (default) APIs.

    Attempts to reload/compile the contract of the specified version, from the .source path.

    Does not clutter up the object() API with names, to avoid shadowing any normal Contract API
    functions starting with letters.

    If a GasOracle is supplied for gas_oracle, we'll use it. Otherwise, we'll fall back to just
    using the gas pricing recommended by our Web3 API provider.
    """

    def __init__(
        self,
        w3_provider,
        agent				= None,		# The Ethereum account of the agent accessing the Contract (if Gas required)
        agent_prvkey: Optional[bytes]	= None,		# Can only query public data, view methods without
        source: Optional[Path]		= None,
        version: Optional[str]		= None,
        name: Optional[str]		= None,
        address: Optional[str]		= None,
        abi: Optional[Dict]		= None,
        bytecode: Optional[bytes]	= None,
        chain: Optional[Chain]		= None,		# The Etherscan network to get data from (basically, always Ethereum)
        speed: Optional[Speed]		= None,		# The Etherscan Gas pricing selection for speed of transactions
        max_usd_per_gas:Optional[float]	= None,		# If not supplied, we'll set a cap of base fee + 2 x priority fee
        gas_oracle: Optional[GasOracle]	= None,
        gas_oracle_timeout: Optional[Tuple[int,float]] = None,  # None avoids waiting, doesn't check
    ):
        # If a GasOracle is supplied, we'll wait up to gas_oracle_timeout seconds for it to report
        # online.  Give it a tickle here to get it started, while we do other time-consuming stuff.
        self._gas_oracle	= GasOracle() if gas_oracle is None else gas_oracle
        gas_oracle_beg		= timer()
        bool( self._gas_oracle )

        self._w3		= Web3( w3_provider )
        self._agent		= agent
        if self._agent is not None:
            self._w3.eth.default_account = str( self._agent )

        if agent_prvkey:
            account_signing	= eth_account.Account.from_key( '0x' + agent_prvkey )
            assert account_signing.address == agent, \
                f"The agent Ethereum Account private key 0x{agent_prvkey} isn't related to agent account address {agent}"
            self._w3.middleware_onion.add(
                construct_sign_and_send_raw_middleware( account_signing ))

        self._address		= address		# An existing deployed contract

        self.__source		= source		# The Contract source path (compile for abi, or to deploy bytecode)
        self._version		= version
        self._name		= name or self.__class__.__name__
        self._abi		= abi
        self._bytescode		= bytecode
        self._max_usd_per_gas	= max_usd_per_gas       # 1.50/21000 --> up to $1.50 for a standard ETH transfer

        self._contract		= None
        self._compiled		= None
        if not self._abi:
            self._compile()

        # We've done all the stuff that may take some time; now, we may need the GasOracle
        # to be online.  Check, and wait if specified.
        if gas_oracle is not None and gas_oracle_timeout is not None:
            while not self._gas_oracle and timer() - gas_oracle_beg < gas_oracle_timeout:
                time.sleep( gas_oracle_timeout / 10 )
            if not self._gas_oracle:
                log.warning( f"Supplied GasOracle still offline after {timer() - gas_oracle_beg:7.2f}s" )

        if self._address:
            # A deployed contract; update any cached data, etc.
            self._contract	= self._w3.eth.contract( address=self._address, abi=self._abi )
            self._update()
        else:
            # Not yet deployed; lets use the provided or compiled ABI and bytecode
            assert self._bytecode, \
                "Must provide abi and bytecode, or Contract source"
            self._contract	= self._w3.eth.contract( abi=self._abi, bytecode=self._bytecode )

    @property
    def _source( self ):
        """Search for the named file, relative to the ., ./contracts/ and finally slip39/invoice/contracts/.

        Also used for caching artifacts related to that source file.

        """
        path			= Path( self.__source or self._name + '.sol' )
        if not path.is_absolute():
            for base in '.', 'contracts', solcx_options['base_path'] / 'contracts':
                check		= Path( base ) / path
                if check.exists():
                    path	= check
                    break
        if self._version is None:
            with open( path, 'rb' ) as f:
                self._version	= sha256( f.read() ).hexdigest()[:6].upper()
        log.info( f"Found {self._name} v{self._version}: {path}" )
        return path

    def _compile( self ):
        """Compiled or reload existing compiled contract.

        See if we've got this code compiled already, w/ the source code version, target solc
        compiler version, and optimization level

            ...sol-abc123-v0.8.17-o100

        """
        source			= self._source
        specs			= [
            '.sol',
            self._version,
            f"v{solc_version}",
        ]
        if {'optimize', 'optimize_runs'} <= set(solcx_options.keys()) and solcx_options['optimize']:
            specs.append( f"o{solcx_options['optimize_runs']}" )
        compiled		= source.with_suffix( '-'.join( specs ))
        if compiled.exists():
            self._compiled	= json.loads( compiled.read_text() )
            log.info( f"{self._name} Reloaded: {compiled}: {json.dumps( self._compiled, indent=4 )}" )
        else:
            self.compiled	= solcx.compile_files(
                source,
                output_values	= ['abi', 'bin'],
                solc_version	= solc_version,
                **solcx_options
            )
            compiled.write_text( json.dumps( self.compiled, indent=4 ))
            log.info( f"{self._name} Compiled: {compiled}: {json.dumps( self.compiled, indent=4 )}" )

        key			= self._abi_key( self._name, compiled )
        self._bytecode		= self._compiled[key]['bin']
        self._abi		= self._compiled[key]['abi']

    def _call( self, name, *args, gas=None, **kwds ):
        """Invoke function name on deployed contract w/ supplied positional args.  For example,
        to call a function we'd normally use:

            self._contract.functions.erc20s_len( ).call({ 'to': "0x", ... })
                                        *args --^       ^-- **kwds

        Since we've already defaulted the self._agent, all we need is the gas price.  But, pass
        any kwds as transaction options.

        The default is to pass 0 gas; must be a free transaction, such as a public value or
        'view' method.  This default will cause any Gas-consuming function call to fail.

        """
        try:
            func		= getattr( self._contract.functions, name )
            # If tx details or a gas limit is supplied, also provide gas pricing; this must be
            # intended to be a transaction to the blockchain.  Otherwise, assume this is a zero-cost
            # call, and provide no gas limit or pricing information.
            tx			= {}
            if kwds or gas:
                tx		= kwds | self._gas_price( gas=gas )
            log.info(  f"Calling {self._name}.{name}( {commas( args )} ) w/ tx: {tx!r}" )
            result		= func( *args ).call( tx )
            success		= True
        except Exception as exc:
            result		= repr( exc )
            success		= False
            raise
        finally:
            log.info( f"Called  {self._name}.{name}( {commas( args )} ) -{'-' if success else 'x'}> {result}" )
        return result

    def __getattr__( self, name ):
        """Assume any unknown attribute (not found in any normal way) .name is assumed to be a call to
        a Contract's API function.  Calls to Contract functions that collide w/ methods must be made via

            .call( name, *args, **kwds )

        All positional args are passed to the Contract API function.

        For functions requiring Gas, must supply a max gas= keyword estimate; any additional
        keywords are assumed to be transaction options.

        """
        def curry( *args, **kwds ):
            return self._call( name, *args, **kwds )
        return curry

    def _abi_key( self, name, path=None ):
        """Determine the ABI key for in .compiled source from 'path', matching contract 'name'"""
        try:
            key,		= ( k for k in self._compiled.keys() if k.endswith( f":{name}" ))
        except Exception as exc:
            raise KeyError( f"Failed to find ABI for Contract {name} in Solidity compiled from {path or self._source}" ) from exc
        return key

    def _update( self ):
        pass

    def _gas_price(
        self,
        gas,				# Estimated Gas required
        fail_fast	= None,		# If we predict failure due to gas * price, fail now
        max_factor	= None		# How much can Gas price increase before failing Tx?
    ):
        """Establish maxPriorityFeePerGas Gas fees for a transaction, optionally w/ a computed
        maxFeePerGas for the given estimated (max) amount of Gas required by the transaction.

        Gets the latest gas pricing information from the connected network to compute the Priority
        Fee required and the estimated Base Fee per Gas that the next block is likely to consume.
        From this, we can estimate what the likely total Fee per Gas is likely to be, in Wei.

        Then, uses estimated Ethereum (ETH) price and self._max_usd_per_gas to compute our
        maxFeePerGas.  If the estimated total Fee per Gas exceeds this limit, the transaction is
        likely to fail, and we'll issue a warning (or raise an Exception, if fail_fast is truthy).

        Otherwise, we'll return a Gas fee dict to use with the transaction.

        TODO: Use the Ethereum JSON-RPC spec and connect to a local node:

            https://ethereum.github.io/execution-apis/api-documentation/

        """

        # Find out what the network thinks the required Gas fees need to be.  This could be a
        # Testnet like Goerli, with abnormally low Gas prices.  But, for a real Ethereum Mainnet
        # transaction, it will be an estimate of the latest block's gas price estimates for the next
        # block.
        latest			= self._w3.eth.get_block( 'latest' )
        base_fee		= latest['baseFeePerGas']
        max_priority_fee	= self._w3.eth.max_priority_fee
        est_gas_wei		= base_fee + max_priority_fee
        max_gas_wei		= base_fee + max_priority_fee * ( max_factor or 2 )
        gas_info_network	= dict(
            maxPriorityFeePerGas	= max_priority_fee,     # Priority fee we're willing to pay, in Wei
            maxFeePerGas		= max_gas_wei,		# With a cap at double the priority fee
        )
        gas_info		= gas_info_network

        # Now, find out what Etherscan's Gas Oracle thinks (from the real Ethereum mainnet,
        # usually).  We'll always use these, if they're not estimated, because we want to simulate
        # real Gas costs even during testing.  However, if we don't have real, updated
        # readings, we'll fall back to the chain's recommendations.
        spend			= gas * self._max_usd_per_gas if gas and self._max_usd_per_gas else None
        if bool( self._gas_oracle ):
            # OK, whatever GasOracle was provided claims to be online; use it!
            try:
                gas_info_oracle	= self._gas_oracle.maxPriorityFeePerGas( gas=gas, spend=spend, max_factor=max_factor )
            except Exception as exc:
                log.warning( f"Gas Oracle API failed: {exc}; Using network estimate instead {traceback.format_exc() if log.isEnabledFor( logging.DEBUG ) else ''}"  )
            else:
                if max_fee_wei := gas_info_oracle.get( 'maxFeePerGas' ):
                    if max_fee_wei < est_gas_wei:
                        what		= f"Max Fee: {max_fee_wei / GasOracle.GWEI_WEI:,.4f} Gwei/Gas is below likely Fee: {est_gas_wei / GasOracle.GWEI_WEI:,.4f} Gwei/Gas"
                        if fail_fast:
                            raise RuntimeError( what + f"; Failing transaction on {self.__class__.__name__}" )
                        else:
                            log.warning( what )
                gas_info	= gas_info_oracle
        else:
            log.info( "Gas Oracle not updated; using network Gas estimates instead" )

        def gas_price_in_gwei( prices ):
            return {
                k: f"{v / GasOracle.GWEI_WEI:,.4f} Gwei == {v:,} Wei" if 'Gas' in k else v
                for k, v in prices.items()
            }

        log.info(
            f"Contract Transaction Gas Price: {json.dumps( gas_price_in_gwei( gas_info ), indent=4 )}" + (
                f"vs. network estimate: {json.dumps( gas_price_in_gwei( gas_info_network ), indent=4 )}" if gas_info != gas_info_network else ""
            )
        )

        # Finally, if a transaction gas limit was supplied, we are assuming this is a Gas-using
        # transaction -- pass it through as the starting Gas.
        if gas is not None:
            gas_info.update( gas=gas )
        return gas_info

    def _tx_gas_cost( self, tx, receipt ):
        """Compute a transaction's actual gas cost, in Wei.  We must get the block's baseFeePerGas, and
        add our transaction's "tip" maxPriorityFeePerGas.  All prices are in Wei/Gas.

        """
        block			= self._w3.eth.get_block( receipt.blockNumber )  # w/ Web3 Tester, may be None
        tx_idx			= receipt.transactionIndex
        base_fee		= block.baseFeePerGas
        prio_fee		= tx.maxPriorityFeePerGas
        log.debug( f"Block {block.number!r:10} base fee: {base_fee/GasOracle.GWEI_WEI:7.4f}Gwei + Tx #{tx_idx!r:4} prio fee: {prio_fee/GasOracle.GWEI_WEI:7.4f}Gwei" )
        return base_fee + prio_fee

    def _deploy( self, *args, gas=None, **kwds ):
        """Create an instance of Contract, passing args to the constructor, and kwds to the transaction.

        Once deployed, we can _update.  However, the contract will not be available for subsequent
        calls until the block is accepted.

        """
        assert not self._address, \
            f"You already have an instance of Contract {self._name}: {self._address}"
        assert gas, \
            f"You must specify an estimated gas amount to deploy a Contract, found: {gas}"

        cons_hash		= self._contract.constructor( *args ).transact( kwds | self._gas_price( gas=gas ))
        log.info( f"Web3 Construct {self._name} hash: {cons_hash.hex()}" )
        cons_tx			= self._w3.eth.get_transaction( cons_hash )
        log.info( f"Web3 Construct {self._name} tx: {json.dumps( cons_tx, indent=4, default=str )}" )
        cons_receipt		= self._w3.eth.wait_for_transaction_receipt( cons_hash )
        log.info( f"Web3 Construct {self._name} receipt: {json.dumps( cons_receipt, indent=4, default=str )}" )
        assert cons_receipt.status, \
            f"Deployment of contract was not successful; status == {cons_receipt.status}"

        # The Contract was successfully deployed.  Get its address, and provide an interface to it.
        self._address		= cons_receipt.contractAddress
        self._contract		= self._w3.eth.contract( address=self._address, abi=self._abi )
        gas_cost		= self._tx_gas_cost( cons_tx, cons_receipt )
        log.warning( f"Web3 Construct {self._name} Contract: {len(self._bytecode)} bytes, at Address: {self._address}" )
        log.info( "Web3 Construct {} Gas Used: {} == {:7.4f}Gwei == USD${:9,.2f} ({}): ${:.6f}/byte".format(
            self._name,
            cons_receipt.gasUsed,
            cons_receipt.gasUsed * gas_cost / GasOracle.GWEI_WEI,
            cons_receipt.gasUsed * gas_cost * self._gas_oracle.ETH_USD / GasOracle.ETH_WEI, self._gas_oracle or 'estimated',
            cons_receipt.gasUsed * gas_cost * self._gas_oracle.ETH_USD / GasOracle.ETH_WEI / len( self._bytecode ),
        ))

        # Wait for the next block to be mined, to ensure contract is available for use.  Is this
        # necessary?  Once the Contract receipt is available, perhaps this means that the block has
        # been mined and is ready for API access.  Otherwise, use '<=' here to wait for next block.
        beg			= timer()
        while ( block_number := self._w3.eth.block_number ) < cons_receipt.blockNumber:
            time.sleep( 1 )
        log.info( f"Waited {timer()-beg:.2f}s for block {block_number} to be mined, vs. Contract block: {cons_receipt.blockNumber}" )

        self._update()


offchainoracle_address		= '0x07D91f5fb9Bf7798734C3f606dB065549F6893bb'
offchainoracle_abi		= [{"inputs":[{"internalType":"contract MultiWrapper","name":"_multiWrapper","type":"address"},{"internalType":"contract IOracle[]","name":"existingOracles","type":"address[]"},{"internalType":"enum OffchainOracle.OracleType[]","name":"oracleTypes","type":"uint8[]"},{"internalType":"contract IERC20[]","name":"existingConnectors","type":"address[]"},{"internalType":"contract IERC20","name":"wBase","type":"address"}],"stateMutability":"nonpayable","type":"constructor"},{"anonymous":False,"inputs":[{"indexed":False,"internalType":"contract IERC20","name":"connector","type":"address"}],"name":"ConnectorAdded","type":"event"},{"anonymous":False,"inputs":[{"indexed":False,"internalType":"contract IERC20","name":"connector","type":"address"}],"name":"ConnectorRemoved","type":"event"},{"anonymous":False,"inputs":[{"indexed":False,"internalType":"contract MultiWrapper","name":"multiWrapper","type":"address"}],"name":"MultiWrapperUpdated","type":"event"},{"anonymous":False,"inputs":[{"indexed":False,"internalType":"contract IOracle","name":"oracle","type":"address"},{"indexed":False,"internalType":"enum OffchainOracle.OracleType","name":"oracleType","type":"uint8"}],"name":"OracleAdded","type":"event"},{"anonymous":False,"inputs":[{"indexed":False,"internalType":"contract IOracle","name":"oracle","type":"address"},{"indexed":False,"internalType":"enum OffchainOracle.OracleType","name":"oracleType","type":"uint8"}],"name":"OracleRemoved","type":"event"},{"anonymous":False,"inputs":[{"indexed":True,"internalType":"address","name":"previousOwner","type":"address"},{"indexed":True,"internalType":"address","name":"newOwner","type":"address"}],"name":"OwnershipTransferred","type":"event"},{"inputs":[{"internalType":"contract IERC20","name":"connector","type":"address"}],"name":"addConnector","outputs":[],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"contract IOracle","name":"oracle","type":"address"},{"internalType":"enum OffchainOracle.OracleType","name":"oracleKind","type":"uint8"}],"name":"addOracle","outputs":[],"stateMutability":"nonpayable","type":"function"},{"inputs":[],"name":"connectors","outputs":[{"internalType":"contract IERC20[]","name":"allConnectors","type":"address[]"}],"stateMutability":"view","type":"function"},{"inputs":[{"internalType":"contract IERC20","name":"srcToken","type":"address"},{"internalType":"contract IERC20","name":"dstToken","type":"address"},{"internalType":"bool","name":"useWrappers","type":"bool"}],"name":"getRate","outputs":[{"internalType":"uint256","name":"weightedRate","type":"uint256"}],"stateMutability":"view","type":"function"},{"inputs":[{"internalType":"contract IERC20","name":"srcToken","type":"address"},{"internalType":"bool","name":"useSrcWrappers","type":"bool"}],"name":"getRateToEth","outputs":[{"internalType":"uint256","name":"weightedRate","type":"uint256"}],"stateMutability":"view","type":"function"},{"inputs":[],"name":"multiWrapper","outputs":[{"internalType":"contract MultiWrapper","name":"","type":"address"}],"stateMutability":"view","type":"function"},{"inputs":[],"name":"oracles","outputs":[{"internalType":"contract IOracle[]","name":"allOracles","type":"address[]"},{"internalType":"enum OffchainOracle.OracleType[]","name":"oracleTypes","type":"uint8[]"}],"stateMutability":"view","type":"function"},{"inputs":[],"name":"owner","outputs":[{"internalType":"address","name":"","type":"address"}],"stateMutability":"view","type":"function"},{"inputs":[{"internalType":"contract IERC20","name":"connector","type":"address"}],"name":"removeConnector","outputs":[],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"contract IOracle","name":"oracle","type":"address"},{"internalType":"enum OffchainOracle.OracleType","name":"oracleKind","type":"uint8"}],"name":"removeOracle","outputs":[],"stateMutability":"nonpayable","type":"function"},{"inputs":[],"name":"renounceOwnership","outputs":[],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"contract MultiWrapper","name":"_multiWrapper","type":"address"}],"name":"setMultiWrapper","outputs":[],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"address","name":"newOwner","type":"address"}],"name":"transferOwnership","outputs":[],"stateMutability":"nonpayable","type":"function"}]  # noqa: E501

ierc20metadata_abi		= [
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "owner",
                "type": "address"
            },
            {
                "indexed": True,
                "internalType": "address",
                "name": "spender",
                "type": "address"
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "value",
                "type": "uint256"
            }
        ],
        "name": "Approval",
        "type": "event"
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "from",
                "type": "address"
            },
            {
                "indexed": True,
                "internalType": "address",
                "name": "to",
                "type": "address"
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "value",
                "type": "uint256"
            }
        ],
        "name": "Transfer",
        "type": "event"
    },
    {
        "inputs": [
            {
                "internalType": "address",
                "name": "owner",
                "type": "address"
            },
            {
                "internalType": "address",
                "name": "spender",
                "type": "address"
            }
        ],
        "name": "allowance",
        "outputs": [
            {
                "internalType": "uint256",
                "name": "",
                "type": "uint256"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {
                "internalType": "address",
                "name": "spender",
                "type": "address"
            },
            {
                "internalType": "uint256",
                "name": "amount",
                "type": "uint256"
            }
        ],
        "name": "approve",
        "outputs": [
            {
                "internalType": "bool",
                "name": "",
                "type": "bool"
            }
        ],
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "inputs": [
            {
                "internalType": "address",
                "name": "account",
                "type": "address"
            }
        ],
        "name": "balanceOf",
        "outputs": [
            {
                "internalType": "uint256",
                "name": "",
                "type": "uint256"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [],
        "name": "decimals",
        "outputs": [
            {
                "internalType": "uint8",
                "name": "",
                "type": "uint8"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [],
        "name": "name",
        "outputs": [
            {
                "internalType": "string",
                "name": "",
                "type": "string"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [],
        "name": "symbol",
        "outputs": [
            {
                "internalType": "string",
                "name": "",
                "type": "string"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [],
        "name": "totalSupply",
        "outputs": [
            {
                "internalType": "uint256",
                "name": "",
                "type": "uint256"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {
                "internalType": "address",
                "name": "to",
                "type": "address"
            },
            {
                "internalType": "uint256",
                "name": "amount",
                "type": "uint256"
            }
        ],
        "name": "transfer",
        "outputs": [
            {
                "internalType": "bool",
                "name": "",
                "type": "bool"
            }
        ],
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "inputs": [
            {
                "internalType": "address",
                "name": "from",
                "type": "address"
            },
            {
                "internalType": "address",
                "name": "to",
                "type": "address"
            },
            {
                "internalType": "uint256",
                "name": "amount",
                "type": "uint256"
            }
        ],
        "name": "transferFrom",
        "outputs": [
            {
                "internalType": "bool",
                "name": "",
                "type": "bool"
            }
        ],
        "stateMutability": "nonpayable",
        "type": "function"
    }
]

#
# Some tokens such as Maker(MKR) implement ERC-20 w/o the standard, optional symbols/name; they use
# bytes32 (and uint256 for digits), eg.:
#
#    https://etherscan.io/address/0x9f8f72aa9304c8b593d555f12ef6589cc3a579a2#code
#
ierc20bytes32_abi		= [
    {"constant":True,"inputs":[],"name":"name",    "outputs":[{"name":"","type":"bytes32"}],"payable":False,"stateMutability":"view","type":"function"},
    {"constant":True,"inputs":[],"name":"decimals","outputs":[{"name":"","type":"uint256"}],"payable":False,"stateMutability":"view","type":"function"},
    {"constant":True,"inputs":[],"name":"symbol",  "outputs":[{"name":"","type":"bytes32"}],"payable":False,"stateMutability":"view","type":"function"},
]


#
# An example of one way to cache Web3 connections
#
# WARNING:
# - Not thread-safe.  These cached providers need to be used by a single thread
# - Caches results w/ differing use_provider=, for the same w3_url
# - We do not normalize addresses; use Ethereum addresses that have corect checksums
#
@memoize( maxage=None, maxsize=None, log_at=logging.INFO )
def w3_provider( w3_url, use_provider ):
    """Return a Web3.*Provider associated with the specified URL, optionally using the specified
    provider."""
    if use_provider is None:
        use_provider		= dict(
            wss		= Web3.WebsocketProvider,
            http	= Web3.HTTPProvider,
            https	= Web3.HTTPProvider,
        )[w3_url.split( ':', 1 )[0].lower()]
    return use_provider( w3_url )


@dataclass( eq=True, frozen=True )      # Makes it hashable
class TokenInfo:
    """Represents a Crypto-currency or on-chain Token (eg. ERC-20)"""
    symbol: str
    name: str
    decimals: int
    contract: Optional[str] = None      # If not an ERC-20 token contract, no contract address
    icon: Optional[Union[str,Path]] = None


def tokeninfo( token, chain=None, w3_url=None, use_provider=None ):
    """Lookup or query the specified token for the specified chain w/ the specified API w3_url, by
    symbol, name, un-normalized or normalized (w/ valid checksum) address.  The first time a Token
    is looked up, its contract address must be used; subsequently, a case-insensitive symbol
    (upper-case) or name (lower-case) may be used.  Tokens with names/symbols identical except for
    case are disambiguated by upper/lower case correction.  Thus, two tokens may have a 'name'
    identical to another's 'symbol', yet not collide.

    Loads the known (top few hundred) Tokens for the chain on first call, to:
    - Avoid needing to reach out to the Ethereum blockchain for well-known tokens
    - To check for "trojan" tokens trying to impersonate well-known tokens on a chain

    This is subtle; since we can look up tokens on various blockchains, and the "same" token
    (eg. USDC, WEENUS) can appear on various blockchains with different contract addresses, we don't
    want to have collisions.  So, each chain must have its own cache of contract/symbol/names.

    Since we are already caching here, we don't use memoize; any w3_url targeting the same chain is
    considered identical.

    """
    try:
        token			= Web3.to_checksum_address( token )
    except ValueError:
        pass
    if chain and isinstance( chain, str ):  # The blockchain; eg. "Ethereum", "Goerli"
        chain,			= ( c for c in Chain if c.name.lower() == chain.lower() )
    assert isinstance( chain, (Chain, type(None)) )
    if chain is None:
        chain			= Chain.Ethereum
    if w3_url is None:
        w3_url			= alchemy_url( chain  )

    def aliasinfo( info ):
        cont			= info.contract
        assert cont not in tokeninfo.ERC20s[chain], \
            f"Duplicate ERC-20 contract address {info.contract!r} on {chain}"
        symb			= info.symbol.upper()
        assert symb not in tokeninfo.ERC20s[chain], \
            f"Duplicate ERC-20 symbol {info.symbol!r} on {chain}"
        name			= info.name.lower()
        assert name not in tokeninfo.ERC20s[chain], \
            f"Duplicate ERC-20 name {info.name!r} on {chain}"
        tokeninfo.ERC20s[chain][cont] = info
        tokeninfo.ERC20s[chain][symb] = info
        tokeninfo.ERC20s[chain][name] = info
        return info

    # If token is known in ./Tokens-<chain>.json, return it.  Iff the stated icon file exists,
    # include its Path, else None.  Defaults to an empty dict for each chain.
    if not tokeninfo.ERC20s.setdefault( chain, {} ):
        here			= Path( __file__ ).resolve().parent
        here_json		= here / f"Tokens-{chain.name}.json"
        try:
            loaded		= -1
            if here_json.exists():
                with open( here_json, 'r' ) as json_f:
                    for loaded,info in enumerate( json.loads( json_f.read() )):
                        if icon := info.get( 'icon' ):
                            icon_p	= here / icon
                            if icon_p.exists():
                                info['icon'] = icon_p.resolve()
                            else:
                                info['icon'] = None
                                log.warning( f"ERC-20 token {info['symbol']} references non-existent icon {icon_p}" )
                        aliasinfo( TokenInfo( **info ))
        except Exception as exc:
            log.warning( f"Failed to load known ERC-20 tokens for {chain} from {here_json}: {exc}" )
        else:
            log.warning( f"Loaded {loaded+1} known ERC-20 tokens for {chain} from {here_json}" )

    # Either the contract address w/ checksum, symbol (upper-case) or name (lower-case) may be
    # present in the chain's cache.
    for t in (token, token.upper(), token.lower(), INVOICE_PROXIES.get( token.upper() ), INVOICE_PROXIES.get( token.lower() )):
        if info := tokeninfo.ERC20s[chain].get( t ):
            return info

    # Not (yet) a known token for chain; we'll have to query the blockchain for this token's data.
    # It must, therefore, be a normalized contract address.
    assert Web3.is_checksum_address( token ), \
        f"Cannot query ERC-20 token contract address: {token}; must be valid address w/checksum"
    w3				= Web3( w3_provider( w3_url, use_provider ))
    token_ierc20metadata	= w3.eth.contract( address=token, abi=ierc20metadata_abi )
    try:
        decimals		= token_ierc20metadata.functions.decimals().call()
        symbol			= token_ierc20metadata.functions.symbol().call()
        name			= token_ierc20metadata.functions.name().call()
    except Exception:
        # Hmm.  Some top-100 tokens fail to respond correctly to the standard IERC20MetaData API.
        # Try another commonly used API.  If this raises an Exception, let it through.
        token_ierc20bytes32	= w3.eth.contract( address=token, abi=ierc20bytes32_abi )
        decimals		= token_ierc20bytes32.functions.decimals().call()
        symbol			= token_ierc20bytes32.functions.symbol().call().strip( b'\0' ).decode( 'utf-8' )
        name			= token_ierc20bytes32.functions.name().call().strip( b'\0' ).decode( 'utf-8' )

    # Remember this token; this will fail if it collides w/ a "known" token.
    return aliasinfo( TokenInfo(
        name		= name,
        symbol		= symbol,
        decimals	= decimals,
        contract	= token,
        icon		= None,
    ))
tokeninfo.ERC20s		= {}  # noqa: E305; { <chain>: {'contract': <info>, 'name': <info>, 'symbol': info, ... }}


def tokenknown( name, decimals=None ):
    """If name is a recognized core supported Cryptocurrency, return a TokenInfo useful for formatting.

    Since Bitcoin (and other similar cryptocurrencies) are typically assumed to have 8 decimal
    places (a Sat(oshi) is 1/10^8 of a Bitcoin), we'll make the default decimals//3 precision work
    out to 8).

    """
    try:
        symbol			= Account.supported( name )
    except ValueError as exc:
        log.info( f"Failed to identify currency {name!r} as a supported Cryptocurrency: {exc}" )
        raise
    return TokenInfo(
        symbol		= symbol,
        name		= Account.CRYPTO_SYMBOLS[symbol],
        decimals	= Account.CRYPTO_DECIMALS[symbol] if decimals is None else decimals,
        icon		= next( ( Path( __file__ ).resolve().parent / "Cryptos" ).glob( symbol + '*.*' ), None ),
    )


@memoize( maxage=TOKPRICES_MEMO_MAXAGE, maxsize=TOKPRICES_MEMO_MAXSIZE, log_at=logging.INFO )
def tokenprice( w3_url, chain, token, base, use_wrappers=None, use_provider=None ):
    """Return memoized token address prices, in terms of a base token address.  The resultant Fraction
    is the ratio token/base (if token is greater in value than base, the ratio will be > 1).  If
    None supplied for base, uses Ethereum.

    Queries and returns the tokens' TokenInfo and the price ratio as a Fraction.

    """
    w3				= Web3( w3_provider( w3_url, use_provider ))
    if use_wrappers is None:
        use_wrappers		= True

    token_info			= tokeninfo( token, w3_url=w3_url, chain=chain, use_provider=use_provider )

    offchainoracle_contract	= w3.eth.contract( address=offchainoracle_address, abi=offchainoracle_abi)
    if base is None:
        base_info		= TokenInfo(
            name	= "Ethereum",
            symbol	= "ETH",
            decimals	= 18,
        )
        token_price		= offchainoracle_contract.functions.getRateToEth( token_info.contract, use_wrappers ).call()
    else:
        base_info		= tokeninfo( base, w3_url=w3_url, chain=chain, use_provider=use_provider )
        token_price		= offchainoracle_contract.functions.getRate(
            token_info.contract, base_info.contract, use_wrappers,
        ).call()
    price			= Fraction( token_price, 10 ** 18 )
    price		       *= Fraction( 10 ** token_info.decimals, 10 ** base_info.decimals )
    return token_info, base_info, price


#
# Public APIs for token{infos,prices,ratio}
#
# Defaults to Alchemy API on Ethereum MainNet chain, searches/normalizes token addresses (from
# eg. ./Tokens-Ethereum.json).
#
def tokeninfos( *tokens, **kwds ):
    for token in tokens:
        yield tokeninfo( token, **kwds )


def tokenprices( *tokens, chain=None, w3_url=None, base=None, use_provider=None, use_wrappers=True ):
    """Return a sequence of token prices as a Fraction, optionally in terms of a base token (ETH is
    default).  Will default to use the Ethereum blockchain via the Alchemy API.

    TODO: should use MultiCallContract.multicall to aggregate several getRate... calls into one
    invocation: https://github.com/1inch/spot-price-aggregator/blob/master/examples/multiple-prices.js

    """
    if chain and isinstance( chain, str ):
        chain,			= ( c for c in Chain if c.name.lower() == chain.lower() )
    assert isinstance( chain, (Chain, type(None)) )
    if chain is None:
        chain			= Chain.Ethereum
    if w3_url is None:
        w3_url			= alchemy_url( chain )
    for token in tokens:
        token_addr		= tokeninfo( token, w3_url=w3_url, chain=chain, use_provider=use_provider ).contract
        base_addr		= None
        if base:
            base_addr		= tokeninfo( base,  w3_url=w3_url, chain=chain, use_provider=use_provider ).contract
        yield tokenprice( w3_url, chain, token_addr, base_addr, use_wrappers=use_wrappers, use_provider=use_provider )


def tokenratio( t1, t2, **kwds):
    """Find the price ratio of tokens t1/t2, optionally relative to a certain base token (should be
    irrelevant, as the default will be the main token in the chain, eg. ETH for the Ethereum MainNet.

    """
    (t1_i,_,t1_p),(t2_i,_,t2_p)	= tokenprices( t1, t2, **kwds )
    return t1_i,t2_i,t1_p/t2_p
